# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utility functions for the graph_editor.
"""

from __future__ import annotations

import re
from six import iteritems
import typing
import tensorflow as tf
from tensorflow.python.framework import ops as tf_ops
from tensorflow.python.ops import array_ops as tf_array_ops

__all__ = [
    "make_list_of_op",
    "get_tensors",
    "make_list_of_t",
    "get_generating_ops",
    "get_consuming_ops",
    "ControlOutputs",
    "placeholder_name",
    "make_placeholder_from_tensor",
    "make_placeholder_from_dtype_and_shape",
]


# The graph editor sometimes need to create placeholders, they are named
# "geph_*". "geph" stands for Graph-Editor PlaceHolder.
_DEFAULT_PLACEHOLDER_PREFIX = "geph"


def concatenate_unique(la, lb):
    """Add all the elements of `lb` to `la` if they are not there already.

    The elements added to `la` maintain ordering with respect to `lb`.

    Args:
      la: List of Python objects.
      lb: List of Python objects.
    Returns:
      `la`: The list `la` with missing elements from `lb`.
    """
    la_set = set(la)
    for l in lb:
        if l not in la_set:
            la.append(l)
            la_set.add(l)
    return la


# TODO(fkp): very generic code, it should be moved in a more generic place.
class ListView:
    """Immutable list wrapper.

    This class is strongly inspired by the one in tf.Operation.
    """

    def __init__(self, list_):
        if not isinstance(list_, list):
            raise TypeError("Expected a list, got: {}.".format(type(list_)))
        self._list = list_

    def __iter__(self):
        return iter(self._list)

    def __len__(self):
        return len(self._list)

    def __bool__(self):
        return bool(self._list)

    # Python 3 wants __bool__, Python 2.7 wants __nonzero__
    __nonzero__ = __bool__

    def __getitem__(self, i):
        return self._list[i]

    def __add__(self, other):
        if not isinstance(other, list):
            other = list(other)
        return list(self) + other


# TODO(fkp): very generic code, it should be moved in a more generic place.
def is_iterable(obj):
    """Return true if the object is iterable."""
    if isinstance(obj, tf.Tensor):
        return False
    try:
        _ = iter(obj)
    except Exception:  # pylint: disable=broad-except
        return False
    return True


def flatten_tree(tree, leaves=None):
    """Flatten a tree into a list.

    Args:
      tree: iterable or not. If iterable, its elements (child) can also be
        iterable or not.
      leaves: list to which the tree leaves are appended (None by default).
    Returns:
      A list of all the leaves in the tree.
    """
    if leaves is None:
        leaves = []
    if isinstance(tree, dict):
        for _, child in iteritems(tree):
            flatten_tree(child, leaves)
    elif is_iterable(tree):
        for child in tree:
            flatten_tree(child, leaves)
    else:
        leaves.append(tree)
    return leaves


def transform_tree(tree, fn, iterable_type=tuple):
    """Transform all the nodes of a tree.

    Args:
      tree: iterable or not. If iterable, its elements (child) can also be
        iterable or not.
      fn: function to apply to each leaves.
      iterable_type: type use to construct the resulting tree for unknown
        iterable, typically `list` or `tuple`.
    Returns:
      A tree whose leaves has been transformed by `fn`.
      The hierarchy of the output tree mimics the one of the input tree.
    """
    if is_iterable(tree):
        if isinstance(tree, dict):
            res = tree.__new__(type(tree))
            res.__init__((k, transform_tree(child, fn)) for k, child in iteritems(tree))
            return res
        elif isinstance(tree, tuple):
            # NamedTuple?
            if hasattr(tree, "_asdict"):
                res = tree.__new__(type(tree), **transform_tree(tree._asdict(), fn))
            else:
                res = tree.__new__(type(tree), (transform_tree(child, fn) for child in tree))
            return res
        elif isinstance(tree, typing.Sequence):
            res = tree.__new__(type(tree))
            res.__init__(transform_tree(child, fn) for child in tree)
            return res
        else:
            return iterable_type(transform_tree(child, fn) for child in tree)
    else:
        return fn(tree)


def check_graphs(*args):
    """Check that all the element in args belong to the same graph.

    Args:
      *args: a list of object with a obj.graph property.
    Raises:
      ValueError: if all the elements do not belong to the same graph.
    """
    graph = None
    for i, sgv in enumerate(args):
        if graph is None and sgv.graph is not None:
            graph = sgv.graph
        elif sgv.graph is not None and sgv.graph is not graph:
            raise ValueError("Argument[{}]: Wrong graph!".format(i))


def get_unique_graph(tops, check_types=None, none_if_empty=False):
    """Return the unique graph used by the all the elements in tops.

    Args:
      tops: list of elements to check (usually a list of tf.Operation and/or
        tf.Tensor). Or a tf.Graph.
      check_types: check that the element in tops are of given type(s). If None,
        the types (tf.Operation, tf.Tensor) are used.
      none_if_empty: don't raise an error if tops is an empty list, just return
        None.
    Returns:
      The unique graph used by all the tops.
    Raises:
      TypeError: if tops is not a iterable of tf.Operation.
      ValueError: if the graph is not unique.
    """
    if isinstance(tops, tf_ops.Graph):
        return tops
    if not is_iterable(tops):
        raise TypeError("{} is not iterable".format(type(tops)))
    if check_types is None:
        check_types = (tf.Operation, tf.Tensor)
    elif not is_iterable(check_types):
        check_types = (check_types,)
    g = None
    for op in tops:
        if not isinstance(op, check_types):
            raise TypeError(
                "Expected a type in ({}), got: {}".format(", ".join([str(t) for t in check_types]), type(op))
            )
        if g is None:
            g = op.graph
        elif g is not op.graph:
            raise ValueError("Operation {} does not belong to given graph".format(op))
    if g is None and not none_if_empty:
        raise ValueError("Can't find the unique graph of an empty list")
    return g


def make_list_of_op(ops, check_graph=True, allow_graph=True, ignore_ts=False):
    """Convert ops to a list of `tf.Operation`.

    Args:
      ops: can be an iterable of `tf.Operation`, a `tf.Graph` or a single
        operation.
      check_graph: if `True` check if all the operations belong to the same graph.
      allow_graph: if `False` a `tf.Graph` cannot be converted.
      ignore_ts: if True, silently ignore `tf.Tensor`.
    Returns:
      A newly created list of `tf.Operation`.
    Raises:
      TypeError: if ops cannot be converted to a list of `tf.Operation` or,
       if `check_graph` is `True`, if all the ops do not belong to the
       same graph.
    """
    if isinstance(ops, tf_ops.Graph):
        if allow_graph:
            return ops.get_operations()
        else:
            raise TypeError("allow_graph is False: cannot convert a tf.Graph.")
    else:
        if not is_iterable(ops):
            ops = [ops]
        if not ops:
            return []
        if check_graph:
            check_types = None if ignore_ts else tf_ops.Operation
            get_unique_graph(ops, check_types=check_types)
        return [op for op in ops if isinstance(op, tf_ops.Operation)]


# TODO(fkp): move this function in tf.Graph?
def get_tensors(graph):
    """get all the tensors which are input or output of an op in the graph.

    Args:
      graph: a `tf.Graph`.
    Returns:
      A list of `tf.Tensor`.
    Raises:
      TypeError: if graph is not a `tf.Graph`.
    """
    if not isinstance(graph, tf_ops.Graph):
        raise TypeError("Expected a graph, got: {}".format(type(graph)))
    ts = []
    for op in graph.get_operations():
        ts += op.outputs
    return ts


def make_list_of_t(ts, check_graph=True, allow_graph=True, ignore_ops=False):
    """Convert ts to a list of `tf.Tensor`.

    Args:
      ts: can be an iterable of `tf.Tensor`, a `tf.Graph` or a single tensor.
      check_graph: if `True` check if all the tensors belong to the same graph.
      allow_graph: if `False` a `tf.Graph` cannot be converted.
      ignore_ops: if `True`, silently ignore `tf.Operation`.
    Returns:
      A newly created list of `tf.Tensor`.
    Raises:
      TypeError: if `ts` cannot be converted to a list of `tf.Tensor` or,
       if `check_graph` is `True`, if all the ops do not belong to the same graph.
    """
    if isinstance(ts, tf_ops.Graph):
        if allow_graph:
            return get_tensors(ts)
        else:
            raise TypeError("allow_graph is False: cannot convert a tf.Graph.")
    else:
        if not is_iterable(ts):
            ts = [ts]
        if not ts:
            return []
        if check_graph:
            check_types = None if ignore_ops else tf.Tensor
            get_unique_graph(ts, check_types=check_types)
        return [t for t in ts if isinstance(t, tf.Tensor)]


def get_generating_ops(ts):
    """Return all the generating ops of the tensors in `ts`.

    Args:
      ts: a list of `tf.Tensor`
    Returns:
      A list of all the generating `tf.Operation` of the tensors in `ts`.
    Raises:
      TypeError: if `ts` cannot be converted to a list of `tf.Tensor`.
    """
    ts = make_list_of_t(ts, allow_graph=False)
    return [t.op for t in ts]


def get_consuming_ops(ts):
    """Return all the consuming ops of the tensors in ts.

    Args:
      ts: a list of `tf.Tensor`
    Returns:
      A list of all the consuming `tf.Operation` of the tensors in `ts`.
    Raises:
      TypeError: if ts cannot be converted to a list of `tf.Tensor`.
    """
    ts = make_list_of_t(ts, allow_graph=False)
    ops = []
    for t in ts:
        for op in t.consumers():
            if op not in ops:
                ops.append(op)
    return ops


class ControlOutputs:
    """The control outputs topology."""

    def __init__(self, graph):
        """Create a dictionary of control-output dependencies.

        Args:
          graph: a `tf.Graph`.
        Returns:
          A dictionary where a key is a `tf.Operation` instance and the
             corresponding value is a list of all the ops which have the key
             as one of their control-input dependencies.
        Raises:
          TypeError: graph is not a `tf.Graph`.
        """
        if not isinstance(graph, tf_ops.Graph):
            raise TypeError("Expected a tf.Graph, got: {}".format(type(graph)))
        self._control_outputs = {}
        self._graph = graph
        self._version = None
        self._build()

    def update(self):
        """Update the control outputs if the graph has changed."""
        if self._version != self._graph.version:
            self._build()
        return self

    def _build(self):
        """Build the control outputs dictionary."""
        self._control_outputs.clear()
        ops = self._graph.get_operations()
        for op in ops:
            for control_input in op.control_inputs:
                if control_input not in self._control_outputs:
                    self._control_outputs[control_input] = []
                if op not in self._control_outputs[control_input]:
                    self._control_outputs[control_input].append(op)
        self._version = self._graph.version

    def get_all(self):
        return self._control_outputs

    def get(self, op):
        """return the control outputs of op."""
        if op in self._control_outputs:
            return self._control_outputs[op]
        else:
            return ()

    @property
    def graph(self):
        return self._graph


def scope_finalize(scope):
    if scope and scope[-1] != "/":
        scope += "/"
    return scope


def scope_dirname(scope):
    slash = scope.rfind("/")
    if slash == -1:
        return ""
    return scope[: slash + 1]


def scope_basename(scope):
    slash = scope.rfind("/")
    if slash == -1:
        return scope
    return scope[slash + 1 :]


def placeholder_name(t=None, scope=None, prefix=_DEFAULT_PLACEHOLDER_PREFIX):
    """Create placeholder name for the graph editor.

    Args:
      t: optional tensor on which the placeholder operation's name will be based
        on
      scope: absolute scope with which to prefix the placeholder's name. None
        means that the scope of t is preserved. "" means the root scope.
      prefix: placeholder name prefix.
    Returns:
      A new placeholder name prefixed by "geph". Note that "geph" stands for
        Graph Editor PlaceHolder. This convention allows to quickly identify the
        placeholder generated by the Graph Editor.
    Raises:
      TypeError: if t is not None or a tf.Tensor.
    """
    if scope is not None:
        scope = scope_finalize(scope)
    if t is not None:
        if not isinstance(t, tf.Tensor):
            raise TypeError("Expected a tf.Tenfor, got: {}".format(type(t)))
        op_dirname = scope_dirname(t.op.name)
        op_basename = scope_basename(t.op.name)
        if scope is None:
            scope = op_dirname

        if op_basename.startswith("{}__".format(prefix)):
            ph_name = op_basename
        else:
            ph_name = "{}__{}_{}".format(prefix, op_basename, t.value_index)

        return scope + ph_name
    else:
        if scope is None:
            scope = ""
        return "{}{}".format(scope, prefix)


def make_placeholder_from_tensor(t, scope=None, prefix=_DEFAULT_PLACEHOLDER_PREFIX):
    """Create a `tf.compat.v1.placeholder` for the Graph Editor.

    Note that the correct graph scope must be set by the calling function.

    Args:
      t: a `tf.Tensor` whose name will be used to create the placeholder (see
        function placeholder_name).
      scope: absolute scope within which to create the placeholder. None means
        that the scope of `t` is preserved. `""` means the root scope.
      prefix: placeholder name prefix.

    Returns:
      A newly created `tf.compat.v1.placeholder`.
    Raises:
      TypeError: if `t` is not `None` or a `tf.Tensor`.
    """
    return tf_array_ops.placeholder(
        dtype=t.dtype, shape=t.get_shape(), name=placeholder_name(t, scope=scope, prefix=prefix)
    )


def make_placeholder_from_dtype_and_shape(dtype, shape=None, scope=None, prefix=_DEFAULT_PLACEHOLDER_PREFIX):
    """Create a tf.compat.v1.placeholder for the Graph Editor.

    Note that the correct graph scope must be set by the calling function.
    The placeholder is named using the function placeholder_name (with no
    tensor argument).

    Args:
      dtype: the tensor type.
      shape: the tensor shape (optional).
      scope: absolute scope within which to create the placeholder. None means
        that the scope of t is preserved. "" means the root scope.
      prefix: placeholder name prefix.

    Returns:
      A newly created tf.placeholder.
    """
    return tf_array_ops.placeholder(dtype=dtype, shape=shape, name=placeholder_name(scope=scope, prefix=prefix))


_INTERNAL_VARIABLE_RE = re.compile(r"^__\w+__$")


def get_predefined_collection_names():
    """Return all the predefined collection names."""
    return [getattr(tf_ops.GraphKeys, key) for key in dir(tf_ops.GraphKeys) if not _INTERNAL_VARIABLE_RE.match(key)]


def find_corresponding_elem(target, dst_graph, dst_scope="", src_scope=""):
    """Find corresponding op/tensor in a different graph.

    Args:
      target: A `tf.Tensor` or a `tf.Operation` belonging to the original graph.
      dst_graph: The graph in which the corresponding graph element must be found.
      dst_scope: A scope which is prepended to the name to look for.
      src_scope: A scope which is removed from the original of `target` name.

    Returns:
      The corresponding tf.Tensor` or a `tf.Operation`.

    Raises:
      ValueError: if `src_name` does not start with `src_scope`.
      TypeError: if `target` is not a `tf.Tensor` or a `tf.Operation`
      KeyError: If the corresponding graph element cannot be found.
    """
    src_name = target.name
    if src_scope:
        src_scope = scope_finalize(src_scope)
        if not src_name.startswidth(src_scope):
            raise ValueError("{} does not start with {}".format(src_name, src_scope))
        src_name = src_name[len(src_scope) :]

    dst_name = src_name
    if dst_scope:
        dst_scope = scope_finalize(dst_scope)
        dst_name = dst_scope + dst_name

    if isinstance(target, tf.Tensor):
        return dst_graph.get_tensor_by_name(dst_name)
    if isinstance(target, tf_ops.Operation):
        return dst_graph.get_operation_by_name(dst_name)
    raise TypeError("Expected tf.Tensor or tf.Operation, got: {}", type(target))


def find_corresponding(targets, dst_graph, dst_scope="", src_scope=""):
    """Find corresponding ops/tensors in a different graph.

    `targets` is a Python tree, that is, a nested structure of iterable
    (list, tupple, dictionary) whose leaves are instances of
    `tf.Tensor` or `tf.Operation`

    Args:
      targets: A Python tree containing `tf.Tensor` or `tf.Operation`
        belonging to the original graph.
      dst_graph: The graph in which the corresponding graph element must be found.
      dst_scope: A scope which is prepended to the name to look for.
      src_scope: A scope which is removed from the original of `top` name.

    Returns:
      A Python tree containin the corresponding tf.Tensor` or a `tf.Operation`.

    Raises:
      ValueError: if `src_name` does not start with `src_scope`.
      TypeError: if `top` is not a `tf.Tensor` or a `tf.Operation`
      KeyError: If the corresponding graph element cannot be found.
    """

    def func(top):
        return find_corresponding_elem(top, dst_graph, dst_scope, src_scope)

    return transform_tree(targets, func)



"""
Conditional logic

https://github.com/rwth-i6/returnn_common/issues/24
"""

from __future__ import annotations
from typing import Any, List, TypeVar, Generic, Callable
from tensorflow.python.util import nest
from returnn.tensor import Tensor, ControlFlowContext
import returnn.frontend as rf
import returnn.tf.frontend_layers as rfl
from . import _utils


T = TypeVar("T")


__all__ = ["Cond", "CondModule"]


class Cond(Generic[T]):
    """
    Conditional branching. Basically behaves like ``if ... else ...``.
    Only one branch will be executed, and the condition needs to be a bool scalar.
    This wraps to :class:`CondLayer` in RETURNN and to ``tf.cond`` in TensorFlow.

    Example::

        with Cond(cond) as cond_obj:
          cond_obj.true = mod_true_case(x)
          cond_obj.false = mod_false_case(x)
          y = cond_obj.result

    Corresponds to::

        if cond:
          y = mod_true_case(x)
        else:
          y = mod_false_case(x)

    The context scope has two states corresponding to the True and False computation branch.
    The initial state is the True branch.
    Assigning ``cond_obj.true`` has the side effect of switching the computation to the False branch.
    """

    def __init__(self, condition: Tensor, *, name: str = "cond"):
        self.condition = condition
        self._entered = False
        self._entered_state = True
        self._true_value = None
        self._true_value_set = False
        self._false_value = None
        self._false_value_set = False
        self._result_value = None
        self.layer_module = CondModule(cond=self)
        self.name_ctx = rfl.Layer(
            module=self.layer_module, suggested_name=name, parent=rfl.Layer.current_ctx(), can_access_children=False
        )
        self.name_ctx.custom_layer_name_scope = ""
        self.true_branch_control_flow_ctx = ControlFlowContext(
            kind=ControlFlowContext.Types.CondTrue,
            outer_ctx=self.name_ctx.control_flow_ctx(),
            identifier=self.name_ctx.get_abs_name(),
        )
        self.true_branch_name_ctx = rfl.Layer(
            module=self.layer_module,
            suggested_name="true",
            parent=self.name_ctx,
            virtual=True,
            can_access_children=False,
            new_control_flow_ctx=self.true_branch_control_flow_ctx,
        )
        self.true_branch_name_ctx.is_subnet = True
        self.true_branch_name_ctx.extend_reserved_names({"output"})
        self.false_branch_control_flow_ctx = ControlFlowContext(
            kind=ControlFlowContext.Types.CondFalse,
            outer_ctx=self.name_ctx.control_flow_ctx(),
            identifier=self.true_branch_control_flow_ctx.identifier,
        )
        self.false_branch_name_ctx = rfl.Layer(
            module=self.layer_module,
            suggested_name="false",
            parent=self.name_ctx,
            virtual=True,
            can_access_children=False,
            new_control_flow_ctx=self.false_branch_control_flow_ctx,
        )
        self.false_branch_name_ctx.is_subnet = True
        self.false_branch_name_ctx.extend_reserved_names({"output"})
        self._extra_ops_true_branch: List[Tensor] = []
        self._extra_ops_false_branch: List[Tensor] = []
        self._false_branch_prehooks: List[Callable[[], Any]] = []
        self._false_branch_posthooks: List[Callable[[], Any]] = []

    def __repr__(self):
        return f"Cond{self.name_ctx}"

    def __enter__(self):
        assert not self._entered, f"{self} cannot enter twice"
        self._entered = True
        self._entered_state = True
        self.true_branch_name_ctx.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # First exit any scopes and do cleanup without throwing any exceptions.
        if self._entered:
            if self._true_value is None:
                self.true_branch_name_ctx.__exit__(exc_type, exc_val, exc_tb)
            elif self._false_value is None:
                self.false_branch_name_ctx.__exit__(exc_type, exc_val, exc_tb)
        if not exc_type:  # only do error checking if there was no other exception
            assert self._entered
            assert self._true_value is not None, f"{self} you need to call else_()"
            assert self._false_value is not None, f"{self} you need to call end()"
        self._entered = False

    @property
    def true(self) -> T:
        """
        The getter usually would not be used.
        """
        return self._true_value

    @true.setter
    def true(self, true_value: T):
        """
        Defines the True branch value.
        Enter the False branch.
        Assign self.false afterwards.
        """
        assert self._entered, f"{self} you need to be in the context scope"
        assert self._entered_state is True, f"{self} you cannot enter the else branch twice"
        assert not self._true_value_set
        if isinstance(true_value, Tensor):
            true_value = _utils.copy(true_value, name=self.true_branch_name_ctx.get_child("output"))
        else:
            values_flat = nest.flatten(true_value)  # type: List[Tensor]
            assert values_flat
            for i, v in enumerate(values_flat):
                if v is None:
                    # Dummy value.
                    # Allow this in case the whole output is just None.
                    # Maybe there were other side effects.
                    v = rf.zeros((), dtype="int32")
                assert isinstance(v, Tensor), f"unexpected {true_value!r}, only expects tensors, got {type(v)}"
                if i == 0:
                    values_flat[i] = _utils.copy(v, name=self.true_branch_name_ctx.get_child("output"))
                else:
                    values_flat[i] = _utils.mark_as_output_in_scope(v, scope=self.true_branch_name_ctx)
            if len(values_flat) == 0:
                # maybe there are just ops with side effects, maybe also only in the other branch
                # dummy output
                _utils.copy(rf.zeros((), dtype="int32"), name=self.true_branch_name_ctx.get_child("output"))
            true_value = nest.pack_sequence_as(true_value, values_flat)
        for op in self._extra_ops_true_branch:
            _utils.mark_as_output_in_scope(op, scope=self.true_branch_name_ctx)
        self.true_branch_name_ctx.__exit__(None, None, None)
        self.false_branch_name_ctx.__enter__()
        self.false_branch_name_ctx.extend_reserved_names(
            {v.raw_tensor.name for v in self.true_branch_name_ctx.marked_outputs}
        )
        self._true_value = true_value
        self._true_value_set = True
        self._entered_state = False
        for hook in self._false_branch_prehooks:
            hook()

    @property
    def false(self) -> T:
        """
        The getter usually would not be used.
        """
        return self._false_value

    @false.setter
    def false(self, false_value: T):
        """
        Define the False branch value.
        After this, self.result is available.
        """
        assert self._entered, f"{self} you need to be in the context scope"
        assert (
            self._entered_state is False
        ), f"{self} you need to be in the False branch, have assigned :func:`true` before"
        assert not self._false_value_set
        nest.assert_same_structure(self._true_value, false_value)
        # This needs to match the true() setter logic.
        if isinstance(false_value, Tensor):
            false_value = _utils.copy(false_value, name=self.false_branch_name_ctx.get_child("output"))
        else:
            true_values_flat = nest.flatten(self._true_value)  # type: List[Tensor]
            false_values_flat = nest.flatten(false_value)  # type: List[Tensor]
            assert false_values_flat and len(false_values_flat) == len(true_values_flat)
            for i, (true_v, false_v) in enumerate(zip(true_values_flat, false_values_flat)):
                assert isinstance(true_v, Tensor)
                if false_v is None:  # see above
                    false_v = rf.zeros((), dtype="int32")  # dummy value
                else:
                    assert isinstance(
                        false_v, Tensor
                    ), f"unexpected {false_value!r}, only expects tensors, got {type(false_v)}"
                assert true_v.raw_tensor.parent is self.true_branch_name_ctx
                name = true_v.raw_tensor.name
                assert name not in self.false_branch_name_ctx.children
                false_values_flat[i] = _utils.copy(false_v, name=self.false_branch_name_ctx.get_child(name))
                if name != "output":
                    false_values_flat[i].raw_tensor.layer_dict["is_output_layer"] = True
            if len(false_values_flat) == 0:
                # Same as in true branch, see above.
                _utils.copy(rf.zeros((), dtype="int32"), name=self.true_branch_name_ctx.get_child("output"))
            false_value = nest.pack_sequence_as(false_value, false_values_flat)
        for true_out in self.true_branch_name_ctx.marked_outputs:
            name = true_out.raw_tensor.name
            if name in self.false_branch_name_ctx.children:
                continue
            _utils.zeros_like_as_output_in_scope(true_out, name=self.false_branch_name_ctx.get_child(name))
        for op in self._extra_ops_false_branch:
            op = _utils.mark_as_output_in_scope(op, scope=self.false_branch_name_ctx)
            name = op.raw_tensor.name
            assert name not in self.true_branch_name_ctx.children
            _utils.zeros_like_as_output_in_scope(op, name=self.true_branch_name_ctx.get_child(name))
        self.false_branch_name_ctx.__exit__(None, None, None)
        self._false_value = false_value
        self._false_value_set = True
        for cb in self._false_branch_posthooks:
            cb()
        self._result_value = self.layer_module()

    @property
    def result(self) -> T:
        """
        :return: the result, after you assigned :func:`true` and :func:`false`.
        """
        assert self._true_value_set, f"{self} you need to have defined the true value"
        assert self._false_value_set, f"{self} you need to have defined the false value"
        return self._result_value

    def add_op_to_current_branch(self, op: Tensor):
        """
        :param op: like an assign_op. the value of the tensor is irrelevant, the underlying op is relevant
        """
        assert self._entered, f"{self} you need to be in the context scope"
        (self._extra_ops_true_branch if self._entered_state else self._extra_ops_false_branch).append(op)

    def add_other_branch_prehook(self, callback: Callable[[], Any]):
        """add prehook to the other branch"""
        assert self._entered, f"{self} you need to be in the context scope"
        if not self._entered_state:
            return  # already in false branch, will not enter any other branch
        self._false_branch_prehooks.insert(0, callback)

    def add_other_branch_posthook(self, callback: Callable[[], Any]):
        """add posthook to the other branch"""
        assert self._entered, f"{self} you need to be in the context scope"
        if not self._entered_state:
            return  # already in false branch, will not enter any other branch
        self._false_branch_posthooks.append(callback)


class CondModule(rf.Module):
    """
    This module is used internally by :class:`Cond` to create the RETURNN :class:`CondLayer` for the conditional code.
    This module would not be directly used by the user.
    """

    def __init__(self, cond: Cond):
        super(CondModule, self).__init__()
        self.cond = cond

    def __call__(self):
        """
        Makes layer dict for this loop, i.e. a RecLayer.

        :return: structure like true_value/false_value
        """
        name_ctx = self.cond.name_ctx
        # noinspection PyProtectedMember
        true_value, false_value = self.cond._true_value, self.cond._false_value
        true_values_flat = nest.flatten(true_value)  # type: List[Tensor]
        false_values_flat = nest.flatten(false_value)  # type: List[Tensor]
        assert len(true_values_flat) == len(false_values_flat)
        res = rfl.make_layer(
            {
                "class": "cond",
                "from": [],
                "condition": self.cond.condition,
                "true_layer": {
                    "class": "subnetwork",
                    "from": [],
                    "subnetwork": self.cond.true_branch_name_ctx.make_net(),
                },
                "false_layer": {
                    "class": "subnetwork",
                    "from": [],
                    "subnetwork": self.cond.false_branch_name_ctx.make_net(),
                },
            },
            name=name_ctx,
            predefined_out_data=true_values_flat[0].copy_template()
            if true_values_flat
            else Tensor("dummy", (), "int32"),
        )

        if true_values_flat:
            results = []
            for i, (true_v, false_v) in enumerate(zip(true_values_flat, false_values_flat)):
                assert isinstance(true_v, Tensor) and isinstance(false_v, Tensor)
                true_v: Tensor[rfl.Layer]
                false_v: Tensor[rfl.Layer]
                assert true_v.raw_tensor.parent is self.cond.true_branch_name_ctx
                name = true_v.raw_tensor.name
                if i == 0:
                    results.append(res)
                else:
                    template = true_v.copy_template_set_ctx(name_ctx.control_flow_ctx())
                    # noinspection PyProtectedMember
                    results.append(rfl._get_sub_layer(res, name, data=template))
                results[-1].raw_tensor.layer_extra_dependencies.extend(
                    (self.cond.condition.raw_tensor, true_v.raw_tensor, false_v.raw_tensor)
                )
                true_v.raw_tensor.usages.append(results[-1].raw_tensor)
                false_v.raw_tensor.usages.append(results[-1].raw_tensor)
            res = nest.pack_sequence_as(true_value, results)
            if not results:
                results = [res]
        else:
            results = [res]

        # noinspection PyProtectedMember
        if self.cond._extra_ops_true_branch or self.cond._extra_ops_false_branch:
            # Make sure this is registered as output layer.
            assert not name_ctx.inner_control_flow()  # not implemented
            out = results[0]
            with rf.control_flow_ctx(name_ctx.control_flow_ctx()):
                out = _utils.copy(out, name=name_ctx.root.get_new_child(out.name))
            out.raw_tensor.layer_dict["is_output_layer"] = True
            name_ctx.root.marked_outputs.append(out)

        return res



"""
Dict of tensors.

For example the extern_data in the user config is such a dict,
describing the data coming from the dataset (after batch preparation).

We also might have model_outputs in the user config.
(https://github.com/rwth-i6/returnn/issues/1166)
"""

from __future__ import annotations
from typing import Optional, Union, Any, Type, Dict, Sequence
from .tensor import Tensor
from .dim import Dim


_TensorT = Union[Tensor, Dict[str, Any]]
_DataAutoConvertT = Union[Dict[str, _TensorT], Sequence[_TensorT]]
_DataStrictT = Union[Dict[str, Tensor], Sequence[Tensor]]


class TensorDict:
    """dict of tensors"""

    def __init__(self, data: Optional[_DataStrictT] = None):
        self.data = {}  # type: Dict[str, Tensor]
        if data:
            self.update(data)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.data})"

    def update(self, data: Union[_DataAutoConvertT, TensorDict], *, auto_convert: bool = False):
        """update"""
        if isinstance(data, TensorDict):
            for key, value in data.data.items():
                self.data[key] = value.copy()
        elif isinstance(data, dict):
            for key, value in data.items():
                if auto_convert:
                    value = _convert_to_tensor(value, name=key)
                else:
                    assert isinstance(value, Tensor)
                self.data[key] = value.copy()
        elif isinstance(data, (list, tuple)):
            for value in data:
                if auto_convert:
                    value = _convert_to_tensor(value)
                else:
                    assert isinstance(value, Tensor)
                self.data[value.name] = value.copy()
        else:
            raise TypeError(f"invalid `data` type: {type(data)}")

    def __contains__(self, item: str) -> bool:
        return item in self.data

    def __getitem__(self, item: str) -> Tensor:
        return self.data[item]

    def reset_content(self):
        """reset content, i.e. all raw_tensor's to None, including dyn_size_ext of dim tags"""
        dims = []
        dims_set = set()
        for value in self.data.values():
            value.reset()
            for dim in value.dims:
                if dim not in dims_set:
                    dims_set.add(dim)
                    dims.append(dim)
        for dim in dims:
            dim.reset_batch_and_raw()

    def copy_template(self) -> TensorDict:
        """copy template"""
        return TensorDict({k: v.copy_template() for k, v in self.data.items()})

    def as_raw_tensor_dict(
        self,
        *,
        include_const_sizes: bool = False,
        include_scalar_dyn_sizes: bool = True,
        exclude_duplicate_dims: bool = False,
        expected_value_type: Union[Type, Sequence[Type]] = object,
    ) -> Dict[str, Any]:
        """
        :return: dict of raw tensors, including any sequence lengths / dynamic sizes
        """
        assert not (include_const_sizes and not include_scalar_dyn_sizes)
        visited_dims = set()
        out = {}
        for key, value in self.data.items():
            assert key not in out
            assert isinstance(
                value.raw_tensor, expected_value_type
            ), f"key {key} {value}: unexpected {type(value.raw_tensor)}, expected {expected_value_type}"
            out[key] = value.raw_tensor
            for i, dim in enumerate(value.dims):
                if exclude_duplicate_dims and dim in visited_dims:
                    continue
                key_ = f"{key}:size{i}"
                assert key_ not in out
                if dim.is_batch_dim() and (dim.dyn_size_ext is None or dim.dyn_size_ext.raw_tensor is None):
                    if include_scalar_dyn_sizes:
                        dim_value = dim.get_dim_value()
                        assert isinstance(
                            dim_value, expected_value_type
                        ), f"key {key_} {dim}: unexpected {type(dim_value)}, expected {expected_value_type}"
                        out[key_] = dim_value
                elif dim.dyn_size_ext is not None:
                    if include_scalar_dyn_sizes or dim.dyn_size_ext.dims:
                        assert isinstance(dim.dyn_size_ext.raw_tensor, expected_value_type), (
                            f"key {key_} {dim} {dim.dyn_size_ext}:"
                            f" unexpected {type(dim.dyn_size_ext.raw_tensor)}, expected {expected_value_type}"
                        )
                        out[key_] = dim.dyn_size_ext.raw_tensor
                elif dim.size is not None:
                    if include_scalar_dyn_sizes and include_const_sizes:
                        assert isinstance(
                            dim.size, expected_value_type
                        ), f"key {key_} {dim}: unexpected {type(dim.size)}, expected {expected_value_type}"
                        out[key_] = dim.size
                else:
                    raise Exception(f"cannot handle dim: {dim}")
                visited_dims.add(dim)
        return out

    def assign_from_raw_tensor_dict_(
        self,
        raw_tensor_dict: Dict[str, Any],
        *,
        with_scalar_dyn_sizes: bool = True,
        duplicate_dims_are_excluded: bool = False,
    ):
        """
        :param raw_tensor_dict: dict of raw tensors, including any sequence lengths / dynamic sizes
        :param with_scalar_dyn_sizes: `include_scalar_dyn_sizes` was used in :func:`as_raw_tensor_dict`
        :param duplicate_dims_are_excluded: `exclude_duplicate_dims` was used in :func:`as_raw_tensor_dict`
        """
        visited_dims = set()
        for key, value in self.data.items():
            assert key in raw_tensor_dict, f"key {key} not in raw_tensor_dict {list(raw_tensor_dict.keys())}"
            value.raw_tensor = raw_tensor_dict[key]
            for i, dim in enumerate(value.dims):
                dim: Dim
                if duplicate_dims_are_excluded and dim in visited_dims:
                    continue
                key_ = f"{key}:size{i}"
                dim.reset_raw(only_self=True)
                if dim.is_batch_dim() and dim.dyn_size_ext is None:
                    dim.dyn_size_ext = Tensor("batch", [], dtype="int32")
                if dim.dyn_size_ext is not None:
                    if not with_scalar_dyn_sizes and not dim.dyn_size_ext.dims:
                        pass
                    else:
                        assert key_ in raw_tensor_dict, f"keys: f{raw_tensor_dict.keys()}"
                        dim.dyn_size_ext.raw_tensor = raw_tensor_dict[key_]
                else:
                    if key_ in raw_tensor_dict:
                        assert dim.size == raw_tensor_dict[key_]
                visited_dims.add(dim)


def _convert_to_tensor(opts: _TensorT, *, name: Optional[str] = None) -> Tensor:
    """
    :param opts:
    """
    if isinstance(opts, Tensor):
        return opts
    assert isinstance(opts, dict)
    opts = opts.copy()
    if name:
        opts["name"] = name
    else:
        assert "name" in opts, f"missing `name` in Tensor {opts!r}"
    return Tensor(**opts)



"""
RETURNN frontend (returnn.frontend) tests
"""

from __future__ import annotations
from typing import Tuple
import _setup_test_env  # noqa
import returnn.frontend as rf
from returnn.tensor import Tensor, Dim, TensorDict
from rf_utils import run_model


def test_constant_bool():
    class _Net(rf.Module):
        def __call__(self) -> Tuple[Tensor, Dim]:
            dim = Dim(3, name="dim")
            return rf.constant(False, dims=[dim]), dim

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        extern_data  # noqa
        out, dim = model()
        assert out.dtype == "bool"
        out.mark_as_default_output(shape=[dim])

    run_model(TensorDict(), lambda *, epoch, step: _Net(), _forward_step)



"""
Lion optimizer <https://arxiv.org/abs/2302.06675>

Code adapted from https://github.com/lucidrains/lion-pytorch/,
which is adapted from https://github.com/google/automl/blob/master/lion/lion_pytorch.py.
"""

from __future__ import annotations
from typing import Optional, Tuple, Callable
import inspect
import torch
from torch.optim.optimizer import Optimizer


class Lion(Optimizer):
    """
    Lion (Evolved Sign Momentum (Evo_l_ved S_i_gn M_o_me_n_tum)) optimizer <https://arxiv.org/abs/2302.06675>
    """

    def __init__(
        self,
        params,
        lr: float = 1e-4,
        betas: Tuple[float, float] = (0.9, 0.99),
        weight_decay: float = 0.0,
        use_triton: Optional[bool] = None,
        decoupled_weight_decay: bool = False,
    ):
        assert lr > 0.0
        assert all([0.0 <= beta <= 1.0 for beta in betas])

        self._init_lr = lr
        self.decoupled_wd = decoupled_weight_decay

        defaults = dict(lr=lr, betas=betas, weight_decay=weight_decay)

        super().__init__(params, defaults)

        if use_triton is None:
            use_triton = bool(triton_update_fn)
        self.use_triton = use_triton

    @torch.no_grad()
    def step(self, closure: Optional[Callable] = None):
        """update step"""
        loss = None
        if closure is not None:
            with torch.enable_grad():
                loss = closure()

        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue

                beta1, beta2 = group["betas"]
                grad, lr, wd, state, decoupled_wd, init_lr = (
                    p.grad,
                    group["lr"],
                    group["weight_decay"],
                    self.state[p],
                    self.decoupled_wd,
                    self._init_lr,
                )

                # maybe decoupled weight decay

                if decoupled_wd:
                    wd /= init_lr

                # init state - exponential moving average of gradient values

                if len(state) == 0:
                    state["exp_avg"] = torch.zeros_like(p)

                exp_avg = state["exp_avg"]

                if self.use_triton and p.is_cuda:
                    triton_update_fn(p, grad, exp_avg, lr, wd, beta1, beta2)
                else:
                    update_fn(p, grad, exp_avg, lr, wd, beta1, beta2)

        return loss


# update functions


def update_fn(p, grad, exp_avg, lr, wd, beta1, beta2):
    """
    Lion update function
    """
    # stepweight decay

    p.data.mul_(1.0 - lr * wd)

    # weight update

    update = exp_avg.clone().mul_(beta1).add(grad, alpha=1.0 - beta1).sign_()
    p.add_(update, alpha=-lr)

    # decay the momentum running average coefficient

    exp_avg.mul_(beta2).add_(grad, alpha=1.0 - beta2)


try:
    # noinspection PyPackageRequirements
    import triton

    # noinspection PyPackageRequirements
    import triton.language as tl
except ImportError as e:
    triton = None
    tl = None


# restore_value is not available in older versions of triton
if triton and "restore_value" in inspect.signature(triton.autotune).parameters:
    # triton cuda kernel

    # noinspection PyPep8Naming,PyArgumentList
    @triton.autotune(
        configs=[
            triton.Config({"BLOCK_SIZE": 128}, num_warps=4),
            triton.Config({"BLOCK_SIZE": 1024}, num_warps=8),
        ],
        key=["n_elements"],
        restore_value=["p_ptr", "exp_avg_ptr"],
    )
    @triton.jit
    def _triton_update_fn_kernel(
        p_ptr,
        grad_ptr,
        exp_avg_ptr,
        lr,
        wd,
        beta1,
        beta2,
        n_elements,
        BLOCK_SIZE: tl.constexpr,
    ):
        pid = tl.program_id(axis=0)

        block_start = pid * BLOCK_SIZE
        offsets = block_start + tl.arange(0, BLOCK_SIZE)

        mask = offsets < n_elements

        # offsetted pointers

        offset_p_ptr = p_ptr + offsets
        offset_grad_ptr = grad_ptr + offsets
        offset_exp_avg_ptr = exp_avg_ptr + offsets

        # load

        p = tl.load(offset_p_ptr, mask=mask)
        grad = tl.load(offset_grad_ptr, mask=mask)
        exp_avg = tl.load(offset_exp_avg_ptr, mask=mask)

        # stepweight decay

        p = p * (1 - lr * wd)

        # diff between momentum running average and grad

        diff = exp_avg - grad

        # weight update

        update = diff * beta1 + grad

        # torch.sign

        can_update = update != 0
        update_sign = tl.where(update > 0, -lr, lr)

        p = p + update_sign * can_update

        # decay the momentum running average coefficient

        exp_avg = diff * beta2 + grad

        # store new params and momentum running average coefficient

        tl.store(offset_p_ptr, p, mask=mask)
        tl.store(offset_exp_avg_ptr, exp_avg, mask=mask)

    def triton_update_fn(
        p: torch.Tensor, grad: torch.Tensor, exp_avg: torch.Tensor, lr: float, wd: float, beta1: float, beta2: float
    ):
        """
        Lion update function using triton kernel
        """
        assert all([t.is_cuda for t in (p, grad, exp_avg)])
        n_elements = p.numel()

        def _grid(meta):
            return tuple((triton.cdiv(n_elements, meta["BLOCK_SIZE"]),))

        _triton_update_fn_kernel[_grid](p, grad, exp_avg, lr, wd, beta1, beta2, n_elements)

else:
    triton_update_fn = None



"""
Provides :class:`PostprocessingDataset`.
"""

from __future__ import annotations

from itertools import islice
from numpy.random import RandomState
from typing import Any, Callable, Dict, Iterator, List, Optional, Tuple, TypeVar

from returnn.datasets.basic import DatasetSeq
from returnn.datasets.util.strings import str_to_numpy_array
from returnn.datasets.util.vocabulary import Vocabulary
from returnn.tensor import Tensor, TensorDict
from returnn.tensor.dim import Dim
from returnn.util import basic as util
from .basic import init_dataset
from .cached2 import CachedDataset2

__all__ = ["PostprocessingDataset", "LaplaceOrdering", "Sequential"]


class PostprocessingDataset(CachedDataset2):
    """
    A dataset that allows for generic online post-processing of data from another
    dataset using a function on the segment level and on the level of multiple
    segments via an iterator.

    This allows integrating various data augmentation techniques like e.g. Mixup,
    SpecAugment or speed perturbation into the data loading pipeline.

    The integration into the data loading pipeline makes it easy to distribute the
    data processing work across multiple CPU cores using `MultiProcDataset` and in
    turn frees the GPU from data preprocessing tasks.

    Example usage::

        from returnn.tensor.dim import Dim, DimTypes

        time_dim = Dim(None, kind=DimTypes.Spatial)
        new_data_dim = Dim(128)

        train = {
            "class": "PostprocessingDataset",
            "dataset": {
                "class": "HDFDataset",
                "files": ["/path/to/data.hdf"],
            },
            # one of them, but not both:
            # (data: TensorDict, *, rng: numpy.random.RandomState, **kwargs) -> TensorDict
            "map_seq": map_seq,
            # (iter: Iterator[TensorDict], *, rng: numpy.random.RandomState, **kwargs) -> Iterator[TensorDict]
            "map_seq_stream": map_seqs,
            # only required when data shapes change wrt. the wrapped dataset:
            "map_outputs": {
                "data": {"dims": [time_dim, new_data_dim]},
            },
        }

    The dataset itself does not support its own seq ordering and relies on the wrapped
    dataset for seq ordering instead. Specifying a ``seq_ordering`` other than ``default``
    results in an error.

    However, we provide an iterator that implements the common `laplace:.NUM_SEQS_PER_BIN`-variant
    of seq ordering that any custom ``map_seq_stream``-style postprocessing iterator can be composed
    with to implement the ordering via :class:`LaplaceOrdering`.

    Like this::

        from returnn.datasets.postprocessing import LaplaceOrdering, Sequential

        def my_map_seq_stream(iterator):
            ...

        train = {
            "class": "PostprocessingDataset",
            # ...
            "map_seq_stream": Sequential(
                my_map_seq_stream,
                LaplaceOrdering(num_seqs_per_bin=1000),
            ),
        }
    """

    def __init__(
        self,
        dataset: Dict[str, Any],
        map_seq: Optional[Callable] = None,
        map_seq_stream: Optional[Callable] = None,
        map_outputs: Optional[Dict[str, Any]] = None,
        map_seq_stream_preserves_num_seqs: Optional[bool] = None,
        **kwargs,
    ):
        """
        :param dataset: inner dataset to be post-processed
        :param map_seq: post processor function operating on the single-segment level.
            Signature: `(data: TensorDict, *, rng: numpy.random.RandomState, **kwargs) -> TensorDict`
            To avoid confusion on the order of how the processing functions are applied to the data, only one of
            ``map_seq`` and ``map_seq_stream`` can be specified at a time.
            To ensure forwards compatibility, the function must accept ``**kwargs`` as its last argument.
            This is enforced by passing randomly named parameters at runtime.
        :param map_seq_stream: post processor function operating on the multiple segment level via an iterator.
            Allows merging multiple segments into one, or generating multiple output segments from one input segment.
            Signature:
                ``(iter: Iterator[TensorDict], *, rng: numpy.random.RandomState, **kwargs) -> Iterator[TensorDict]``
            To avoid confusion on the order of how the processing functions are applied to the data, only one of
            ``map_seq`` and ``map_seq_stream`` can be specified at a time.
            To ensure forwards compatibility, the function must accept ``**kwargs`` as its last argument.
            This is enforced by passing randomly named parameters at runtime.
        :param map_outputs: Type and axis specification of the outputs of the mapping functions,
            like extern_data and model_outputs.
            To simplify the common case when no shapes change, this value can be left unspecified. The dataset then
            assumes the same data layout as returned by the wrapped dataset.
            Example: `map_outputs={"data": {"dim": 42}}`
        :param map_seq_stream_preserves_num_seqs: whether the function in map_seq_stream preserves the number of
            sequences, i.e. for every input sequence there is exactly one output sequence.
        :param kwargs: see :class:`CachedDataset2`, :class:`Dataset`
        """
        super().__init__(**kwargs)

        if self.seq_ordering != "default":
            raise ValueError(f"{self}: specify seq_ordering in wrapped dataset, not in {self.__class__.__name__}")
        if map_seq is None and map_seq_stream is None:
            raise ValueError(f"{self}: need to either set map_seq or map_seq_stream")
        if map_seq and map_seq_stream:
            raise ValueError(f"{self}: cannot set both map_seq and map_seq_stream")
        if map_seq and map_seq_stream_preserves_num_seqs is not None:
            raise ValueError(f"{self}: map_seq_stream_preserves_num_seqs is only allowed with map_seq_stream")

        self._dataset_def = dataset
        self._map_seq = map_seq
        self._map_seq_stream = map_seq_stream
        if map_seq_stream_preserves_num_seqs is None and map_seq_stream is not None:
            map_seq_stream_preserves_num_seqs = getattr(map_seq_stream, "preserves_num_seqs", None)
        self._map_seq_stream_preserves_num_seqs = map_seq_stream_preserves_num_seqs or False
        self._map_outputs = map_outputs
        self._rng = RandomState(self._get_random_seed_for_epoch(0))
        self._seq_list_for_validation: Optional[List[str]] = None

        self._dataset = init_dataset(self._dataset_def, parent_dataset=self)
        if self._map_seq_stream is None or self._map_seq_stream_preserves_num_seqs:
            # if the stream mapper is set, the num_seqs may change and the estimation is less accurate
            self._estimated_num_seqs = self._dataset.estimated_num_seqs
        self._data_iter: Optional[Iterator[Tuple[int, TensorDict]]] = None
        self._data_iter_produced_num_seqs = 0

        self._in_tensor_dict_template = TensorDict(
            {name: self._make_tensor_template_from_input(name) for name in self._dataset.get_data_keys()}
        )
        self.labels = {}
        if self._map_outputs is not None:
            self._out_tensor_dict_template = TensorDict()
            self._out_tensor_dict_template.update(self._map_outputs, auto_convert=True)
        else:
            self._out_tensor_dict_template = self._in_tensor_dict_template.copy_template()
            self.labels = self._dataset.labels.copy()
        # update only after _out_tensor_dict_template has been created from _in_tensor_dict_template
        self._in_tensor_dict_template.update({"seq_tag": {"dims": (), "dtype": "string"}}, auto_convert=True)
        self.num_outputs = {
            k: (t.sparse_dim.size if t.sparse_dim else t.shape[-1] if len(t.shape) > 0 else 1, t.ndim)
            for k, t in self._out_tensor_dict_template.data.items()
        }
        self._default_input = "data" if "data" in self.num_outputs else next(iter(self.num_outputs.keys()))
        self.num_inputs = self.num_outputs[self._default_input][0]

        for k, t in self._out_tensor_dict_template.data.items():
            if self.labels.get(k):
                continue
            if t.vocab:
                self.labels[k] = t.vocab.labels
            elif t.sparse_dim:  # sparse_dim but not vocab
                self.labels[k] = list(map(str, range(t.sparse_dim.dimension)))  # dummy labels

    def init_seq_order(
        self, epoch: Optional[int] = None, seq_list: Optional[List[str]] = None, seq_order: Optional[List[int]] = None
    ):
        """
        :param epoch:
        :param seq_list:
        :param seq_order:
        :return: whether the order changed (True is always safe to return)
        """
        super().init_seq_order(epoch=epoch, seq_list=seq_list, seq_order=seq_order)

        if self._map_seq_stream is not None:
            if seq_list is not None:
                raise ValueError("map_seq_stream is set, cannot specify custom seq_list")
            if seq_order is not None:
                raise ValueError("map_seq_stream is set, cannot specify custom seq_order")

        if epoch is None and seq_list is None and seq_order is None:
            self._num_seqs = 0
            return True

        self._rng = RandomState(self._get_random_seed_for_epoch(epoch=epoch))
        assert self._dataset is not None
        self._dataset.init_seq_order(epoch=epoch, seq_list=seq_list, seq_order=seq_order)
        self._data_iter = enumerate(self._build_mapping_iter())
        self._data_iter_produced_num_seqs = 0
        self._seq_list_for_validation = seq_list
        if self._map_seq_stream is None or self._map_seq_stream_preserves_num_seqs:
            # If we don't have an iterable mapper (or the user explicitly specifies this),
            # we know the number of segments exactly equals the number of segments in the wrapped dataset
            try:
                self._num_seqs = self._dataset.num_seqs
            except NotImplementedError:
                pass  # some datasets don't know their num_seqs
        return True

    def get_current_seq_order(self):
        """:return: current seq order of wrapped dataset, if map_seq_stream is not used"""
        if self._map_seq_stream is not None:
            raise Exception(f"{self}: get_current_seq_order is not allowed when map_seq_stream is set.")
        assert self._dataset is not None
        return self._dataset.get_current_seq_order()

    def get_data_keys(self):
        """:return: available data keys"""
        return list(self._out_tensor_dict_template.data.keys())

    def get_data_dtype(self, key):
        """:return: dtype of data entry `key`"""
        return self._out_tensor_dict_template.data[key].dtype

    def supports_sharding(self) -> bool:
        """:return: whether this dataset supports sharding"""
        assert self._dataset is not None
        return self._dataset.supports_sharding()

    def _collect_single_seq(self, seq_idx: int) -> Optional[DatasetSeq]:
        while True:
            try:
                loaded_seq_idx, tensor_dict = next(self._data_iter)
                self._data_iter_produced_num_seqs += 1
                if self._num_seqs is not None:
                    assert self._data_iter_produced_num_seqs <= self._num_seqs, (
                        f"{self}: map_seq_stream yielded more seqs ({self._data_iter_produced_num_seqs}) "
                        f"than expected ({self._num_seqs}). map_seq_stream_preserves_num_seqs is set to "
                        f"{self._map_seq_stream_preserves_num_seqs}"
                    )
            except StopIteration:
                if self._num_seqs is not None:
                    assert self._data_iter_produced_num_seqs == self._num_seqs, (
                        f"{self}: map_seq_stream yielded {self._data_iter_produced_num_seqs} seqs, "
                        f"while {self._num_seqs} were expected. map_seq_stream_preserves_num_seqs is set to "
                        f"{self._map_seq_stream_preserves_num_seqs}"
                    )
                return None
            assert loaded_seq_idx <= seq_idx, "_collect_single_seq must be done monotonically"
            if loaded_seq_idx != seq_idx:
                continue
            seq = DatasetSeq(
                features={k: t.raw_tensor for k, t in tensor_dict.data.items() if k != "seq_tag"},
                seq_idx=seq_idx,
                seq_tag=str(tensor_dict["seq_tag"].raw_tensor),
            )
            return seq

    def _build_mapping_iter(self) -> Iterator[TensorDict]:
        """
        :return: an iterator applying both the segment level and across-segment transformations on the given dataset
        """

        def _validate_tensor_dict_iter(inner: Iterator[TensorDict]) -> Iterator[TensorDict]:
            for t_dict in inner:
                assert "seq_tag" in t_dict.data, "seq_tag dropped from TensorDict in postprocessing pipeline"
                for data_key, out_t in self._out_tensor_dict_template.data.items():
                    in_t = t_dict.data[data_key]
                    assert (
                        in_t.ndim == out_t.batch_ndim
                        and in_t.dtype == out_t.dtype
                        and all(d.dimension in (d_, None) for (d, d_) in zip(in_t.dims, out_t.shape))
                    )
                yield t_dict

        data_iter = self._iterate_dataset()
        if self._map_seq_stream is not None:
            data_iter = self._map_seq_stream(data_iter, epoch=self.epoch, rng=self._rng, **util.get_fwd_compat_kwargs())
            assert isinstance(
                data_iter, Iterator
            ), f"map_seq_stream must produce an {Iterator.__name__}, but produced {type(data_iter).__name__}"
        return _validate_tensor_dict_iter(data_iter)

    def _iterate_dataset(self) -> Iterator[TensorDict]:
        """
        :return: generator providing data samples in the form of a TensorDict
        """
        data_keys = self._dataset.get_data_keys()

        seq_index = 0
        while self._dataset.is_less_than_num_seqs(seq_index):
            self._dataset.load_seqs(seq_index, seq_index + 1)

            tensor_dict = self._in_tensor_dict_template.copy_template()
            for data_key in data_keys:
                tensor_dict.data[data_key].raw_tensor = self._dataset.get_data(seq_index, data_key)
            seq_tag_tensor = str_to_numpy_array(self._dataset.get_tag(seq_index))
            tensor_dict.data["seq_tag"].raw_tensor = seq_tag_tensor

            if self._map_seq is not None:
                tensor_dict = self._map_seq(
                    tensor_dict, epoch=self.epoch, seq_idx=seq_index, rng=self._rng, **util.get_fwd_compat_kwargs()
                )
                assert isinstance(
                    tensor_dict, TensorDict
                ), f"map_seq must produce a {TensorDict.__name__}, but produced {type(tensor_dict).__name__}"

                # Re-adding the seq tag here causes no harm in case it's dropped since we don't
                # add/drop any segments w/ the non-iterator postprocessing function.
                if "seq_tag" not in tensor_dict.data:
                    tensor_dict.data["seq_tag"].raw_tensor = seq_tag_tensor

                if self._seq_list_for_validation is not None:
                    seq_tag = self._seq_list_for_validation[seq_index]
                    tag_of_seq = tensor_dict.data["seq_tag"].raw_tensor.item()
                    assert (
                        tag_of_seq == seq_tag
                    ), f"seq tag mismath: {tag_of_seq} != {seq_tag} for seq index {seq_index} when seq list is given"

            yield tensor_dict
            seq_index += 1

    def _make_tensor_template_from_input(self, data_key: str) -> Tensor:
        dtype = self._dataset.get_data_dtype(data_key)
        if dtype == "string":
            dims = []
        else:
            feature_dims = [
                Dim(dimension=dim, name=f"{data_key}_dim{i + 1}")
                for i, dim in enumerate(self._dataset.get_data_shape(data_key))
            ]
            dims = [Dim(dimension=None, name=f"{data_key}_frame"), *feature_dims]
        sparse_dim = None
        if self._dataset.is_data_sparse(data_key):
            sparse_dim = Dim(dimension=self._dataset.get_data_dim(data_key), name=f"{data_key}_sparse")
            labels = self._dataset.labels.get(data_key)
            if labels and len(labels) == sparse_dim.dimension:
                sparse_dim.vocab = Vocabulary.create_vocab_from_labels(self._dataset.labels[data_key])
        return Tensor(data_key, dims=dims, dtype=dtype, sparse_dim=sparse_dim)


class LaplaceOrdering(Callable[[Iterator[TensorDict]], Iterator[TensorDict]]):
    """
    Iterator compatible with :class:`PostprocessingDataset`'s ``map_seq_stream`` applying
    laplace sequence ordering based on the number of segments per bin.

    To be composed with any custom data postprocessing logic via :class:`Sequential`.
    """

    preserves_num_seqs = True

    def __init__(self, num_seqs_per_bin: int, length_key: str = "data"):
        """
        :param num_seqs_per_bin: number of segments in a single laplace bin.
        :param length_key: data key to determine the segment length from for ordering.
        """
        self.length_key = length_key
        assert num_seqs_per_bin > 0
        self.num_seqs_per_bin = num_seqs_per_bin

    def __call__(self, iterator: Iterator[TensorDict], **kwargs) -> Iterator[TensorDict]:
        """:return: generator applying laplace sequence ordering on the data"""
        iterator = iter(iterator)
        is_down_phase = False

        seq_buffer = list(islice(iterator, self.num_seqs_per_bin))
        has_ended = False
        while True:
            seq_buffer.sort(key=self._get_seq_len, reverse=is_down_phase)

            next_seq_buffer = []

            # Yield items to trainer while gradually pulling more data from PP function.
            # This optimizes CPU load when multiple workers are used.
            for item in seq_buffer:
                yield item

                try:
                    if not has_ended:
                        next_seq_buffer.append(next(iterator))
                except StopIteration:
                    has_ended = True

            if len(seq_buffer) < self.num_seqs_per_bin:
                assert has_ended and not next_seq_buffer
                break

            is_down_phase = not is_down_phase
            seq_buffer = next_seq_buffer

    def _get_seq_len(self, tdict: TensorDict) -> int:
        """
        :return: segment length of the segment in `tdict` as measured by `self.length_key` for comparison.
        """
        return tdict.data[self.length_key].raw_tensor.shape[0]


T = TypeVar("T", TensorDict, Iterator[TensorDict])


class Sequential:
    """
    Callable that composes multiple postprocessing functions into one by sequential application,
    i.e. Sequential(f, g)(x) = (g ∘ f)(x) = g(f(x)).

    Can either compose ``map_seq``-style single-segment processor functions or ``map_seq_stream``-style
    iterators operating on multiple segments. Just make sure not to mix both styles.
    """

    def __init__(self, *postprocessing_funcs: Callable):
        """
        :param postprocessing_funcs: Postprocessing functions to compose.
        """
        self.funcs = postprocessing_funcs

    def __call__(self, arg: T, **kwargs) -> T:
        """:return: result of sequential application of the postprocessing functions"""

        for func in self.funcs:
            arg = func(arg, **kwargs)
        return arg

    @property
    def preserves_num_seqs(self):
        """:return: whether the composed functions all preserve the number of sequences"""
        return all(getattr(f, "preserves_num_seqs", False) for f in self.funcs)



"""
Some generic debugging utilities.
"""

from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Union, Any, Collection, Sequence, Callable, Tuple, List, Dict, TextIO
from types import FunctionType, CodeType, FrameType
import os
import sys
import signal

try:
    import thread
except ImportError:
    import _thread as thread
import threading


if TYPE_CHECKING:
    from returnn.tensor import Tensor, Dim
    import torch


signum_to_signame = {
    k: v for v, k in reversed(sorted(signal.__dict__.items())) if v.startswith("SIG") and not v.startswith("SIG_")
}


global_exclude_thread_ids = set()


def auto_exclude_all_new_threads(func):
    """
    :param T func:
    :return: func wrapped
    :rtype: T
    """

    def wrapped(*args, **kwargs):
        """
        :param args:
        :param kwargs:
        :return:
        """
        # noinspection PyProtectedMember,PyUnresolvedReferences
        old_threads = set(sys._current_frames().keys())
        res = func(*args, **kwargs)
        # noinspection PyProtectedMember,PyUnresolvedReferences
        new_threads = set(sys._current_frames().keys())
        new_threads -= old_threads
        global_exclude_thread_ids.update(new_threads)
        return res

    return wrapped


def dump_all_thread_tracebacks(
    *, exclude_thread_ids: Optional[Collection[int]] = None, exclude_self: bool = False, file: Optional[TextIO] = None
):
    """
    :param exclude_thread_ids:
    :param exclude_self:
    :param file:
    """
    if exclude_thread_ids is None:
        exclude_thread_ids = set()
    if file is None:
        file = sys.stdout

    from returnn.util.better_exchook import print_tb
    import threading

    if exclude_self:
        exclude_thread_ids = set(list(exclude_thread_ids) + [threading.current_thread().ident])

    if hasattr(sys, "_current_frames"):
        print("", file=file)
        threads = {t.ident: t for t in threading.enumerate()}
        # noinspection PyProtectedMember
        for tid, stack in sorted(sys._current_frames().items()):
            # This is a bug in earlier Python versions.
            # https://bugs.python.org/issue17094
            # Note that this leaves out all threads not created via the threading module.
            if tid not in threads:
                continue
            tags = []
            thread_ = threads.get(tid)
            if thread_:
                assert isinstance(thread_, threading.Thread)
                if thread_ is threading.current_thread():
                    tags += ["current"]
                # noinspection PyUnresolvedReferences,PyProtectedMember
                if isinstance(thread_, threading._MainThread):
                    tags += ["main"]
                tags += [str(thread_)]
            else:
                tags += ["unknown with id %i" % tid]
            print("Thread %s:" % ", ".join(tags), file=file)
            if tid in global_exclude_thread_ids:
                print("(Auto-ignored traceback.)", file=file)
            elif tid in exclude_thread_ids:
                print("(Excluded thread.)", file=file)
            else:
                print_tb(stack, file=file)
            print("", file=file)
        print("That were all threads.", file=file)
    else:
        print("Does not have sys._current_frames, cannot get thread tracebacks.", file=file)


def setup_warn_with_traceback():
    """
    Installs some hook for ``warnings.showwarning``.
    """
    import warnings
    from returnn.util.better_exchook import print_tb

    def warn_with_traceback(message, category, filename, lineno, file=None, line=None):
        """
        :param message:
        :param category:
        :param filename:
        :param lineno:
        :param file:
        :param line:
        """
        log = file if hasattr(file, "write") else sys.stderr
        log.write(warnings.formatwarning(message, category, filename, lineno, line))
        # noinspection PyProtectedMember,PyUnresolvedReferences
        print_tb(sys._getframe(), file=log)

    warnings.showwarning = warn_with_traceback


def init_better_exchook():
    """
    Installs our own ``sys.excepthook``, which uses :mod:`better_exchook`,
    but adds some special handling for the main thread.
    """
    from returnn.util.better_exchook import better_exchook
    from returnn.log import log

    def excepthook(exc_type, exc_obj, exc_tb):
        """
        :param exc_type:
        :param exc_obj:
        :param exc_tb:
        """
        file = log.v1 or sys.stdout

        # noinspection PyBroadException
        try:
            # noinspection PyUnresolvedReferences,PyProtectedMember
            is_main_thread = isinstance(threading.current_thread(), threading._MainThread)
        except Exception:  # Can happen at a very late state while quitting.
            if exc_type is KeyboardInterrupt:
                return
        else:
            if is_main_thread:
                if exc_type is KeyboardInterrupt and getattr(sys, "exited", False):
                    # Got SIGINT twice. Can happen.
                    return
                # An unhandled exception in the main thread. This means that we are going to quit now.
                sys.exited = True
        print(
            "Unhandled exception %s in thread %s, proc %i." % (exc_type, threading.current_thread(), os.getpid()),
            file=file,
        )
        if exc_type is KeyboardInterrupt:
            return

        # noinspection PyUnresolvedReferences,PyProtectedMember
        if isinstance(threading.current_thread(), threading._MainThread):
            main_thread_id = thread.get_ident()
            if not isinstance(exc_type, Exception):
                # We are the main thread and we got an exit-exception. This is likely fatal.
                # This usually means an exit. (We ignore non-daemon threads and procs here.)
                # Print the stack of all other threads.
                dump_all_thread_tracebacks(exclude_thread_ids={main_thread_id}, file=file)

        better_exchook(exc_type, exc_obj, exc_tb, file=file)

    sys.excepthook = excepthook

    def threading_excepthook(args, /):
        """
        Thread-specific excepthook to ensure the main thread is killed on unhandled exceptions in sub threads.
        """
        log_out = log.v1 or sys.stdout
        print(
            f"Unhandled exception in thread {threading.current_thread()}, going to interrupt main thread:", file=log_out
        )
        better_exchook(args.exc_type, args.exc_value, args.exc_traceback, autodebugshell=False, file=log_out)
        thread.interrupt_main()

    threading.excepthook = threading_excepthook

    from returnn.util.basic import to_bool

    if os.environ.get("DEBUG_WARN_WITH_TRACEBACK") and to_bool(os.environ.get("DEBUG_WARN_WITH_TRACEBACK")):
        setup_warn_with_traceback()


def format_signum(signum):
    """
    :param int signum:
    :return: string "signum (signame)"
    :rtype: str
    """
    return "%s (%s)" % (signum, signum_to_signame.get(signum, "unknown"))


# noinspection PyUnusedLocal
def signal_handler(signum, frame):
    """
    Prints a message on stdout and dump all thread stacks.

    :param int signum: e.g. signal.SIGUSR1
    :param frame: ignored, will dump all threads
    """
    print("Signal handler: got signal %s" % format_signum(signum))
    dump_all_thread_tracebacks()


def install_signal_handler_if_default(signum, exceptions_are_fatal=False):
    """
    :param int signum: e.g. signal.SIGUSR1
    :param bool exceptions_are_fatal: if True, will reraise any exceptions. if False, will just print a message
    :return: True iff no exception, False otherwise. not necessarily that we registered our own handler
    :rtype: bool
    """
    try:
        if signal.getsignal(signum) == signal.SIG_DFL:
            signal.signal(signum, signal_handler)
        return True
    except Exception as exc:
        if exceptions_are_fatal:
            raise
        print("Cannot install signal handler for signal %s, exception %s" % (format_signum(signum), exc))
    return False


_native_signal_handler_lib_filename = None


def _get_native_signal_handler_lib_filename() -> str:
    """
    :return: path to our native_signal_handler lib. see :func:`install_native_signal_handler`
    """
    global _native_signal_handler_lib_filename
    if _native_signal_handler_lib_filename:
        return _native_signal_handler_lib_filename

    from returnn.util.basic import NativeCodeCompiler
    import textwrap

    native = NativeCodeCompiler(
        base_name="native_signal_handler",
        code_version=1,
        code=textwrap.dedent(
            # Derived from https://github.com/albertz/playground/blob/master/signal_handler.c.
            # language=C++
            """\
            #include <stdio.h>
            #include <execinfo.h>
            #include <signal.h>
            #include <stdlib.h>
            #include <unistd.h>


            // https://github.com/ruby/ruby/blob/bbfd735b887/vm_core.h#L118
            #if defined(NSIG_MAX)           /* POSIX issue 8 */
            # undef NSIG
            # define NSIG NSIG_MAX
            #elif defined(_SIG_MAXSIG)      /* FreeBSD */
            # undef NSIG
            # define NSIG _SIG_MAXSIG
            #elif defined(_SIGMAX)          /* QNX */
            # define NSIG (_SIGMAX + 1)
            #elif defined(NSIG)             /* 99% of everything else */
            # /* take it */
            #else                           /* Last resort */
            # define NSIG (sizeof(sigset_t) * CHAR_BIT + 1)
            #endif


            sig_t old_signal_handler[NSIG];


            void signal_handler(int sig) {
              void *array[16 * 1024];
              size_t size;

              // get void*'s for all entries on the stack
              size = backtrace(array, sizeof(array)/sizeof(array[0]));

              // print out all the frames to stderr
              fprintf(stderr, "Signal handler: signal %d:\\n", sig);
              backtrace_symbols_fd(array, size, STDERR_FILENO);

              // call previous handler
              signal(sig, old_signal_handler[sig]);
              raise(sig);
            }

            void install_signal_handler() {
              old_signal_handler[SIGSEGV] = signal(SIGSEGV, signal_handler);
              old_signal_handler[SIGBUS] = signal(SIGBUS, signal_handler);
              old_signal_handler[SIGILL] = signal(SIGILL, signal_handler);
              old_signal_handler[SIGABRT] = signal(SIGABRT, signal_handler);
              old_signal_handler[SIGFPE] = signal(SIGFPE, signal_handler);
            }
            """
        ),
        is_cpp=False,
    )
    _native_signal_handler_lib_filename = native.get_lib_filename()
    return _native_signal_handler_lib_filename


def install_native_signal_handler(*, reraise_exceptions: bool = False):
    """
    Installs some own custom C signal handler.
    """
    try:
        import ctypes

        # C code: https://github.com/albertz/playground/blob/master/signal_handler.c
        lib = ctypes.CDLL(_get_native_signal_handler_lib_filename())
        lib.install_signal_handler.return_type = None
        lib.install_signal_handler()
        print("Installed native_signal_handler.so.")

    except Exception as exc:
        print("installNativeSignalHandler exception: %s" % exc)
        if reraise_exceptions:
            raise


def install_lib_sig_segfault():
    """
    Installs libSegFault (common on Unix/Linux).
    """
    try:
        os.environ.setdefault("SEGFAULT_SIGNALS", "all")
        import ctypes
        import ctypes.util

        # libSegFault on Unix/Linux, not on MacOSX
        libfn = ctypes.util.find_library("SegFault")
        assert libfn, "libSegFault not found"
        # Nothing more needed than loading it, it will automatically register itself.
        ctypes.CDLL(libfn)
        print("Installed libSegFault.so.")

    except Exception as exc:
        print("installLibSigSegfault exception: %s" % exc)


def init_faulthandler(sigusr1_chain=False):
    """
    Maybe installs signal handlers, SIGUSR1 and SIGUSR2 and others.
    If no signals handlers are installed yet for SIGUSR1/2, we try to install our own Python handler.
    This also tries to install the handler from the fauldhandler module,
    esp for SIGSEGV and others.

    :param bool sigusr1_chain: whether the default SIGUSR1 handler should also be called.
    """
    # Enable libSigSegfault first, so that we can have both,
    # because faulthandler will also call the original sig handler.
    install_native_signal_handler()
    if sys.platform != "win32":
        # In case that sigusr1_chain, we expect that there is already some handler
        # for SIGUSR1, and then this will not overwrite this handler.
        if install_signal_handler_if_default(signal.SIGUSR1):
            # There is already some handler or we installed our own handler now,
            # so in any case, it's safe that we chain then handler.
            sigusr1_chain = True
        # Why not also SIGUSR2... SGE can also send this signal.
        install_signal_handler_if_default(signal.SIGUSR2)
    try:
        import faulthandler
    except ImportError as e:
        print("faulthandler import error. %s" % e)
    else:
        # Only enable if not yet enabled -- otherwise, leave it in its current state.
        if not faulthandler.is_enabled():
            faulthandler.enable()
            if sys.platform != "win32":
                faulthandler.register(signal.SIGUSR1, all_threads=True, chain=sigusr1_chain)


@auto_exclude_all_new_threads
def init_ipython_kernel():
    """
    Runs IPython in some background kernel, where you can connect to.
    """
    # You can remotely connect to this kernel. See the output on stdout.
    try:
        # noinspection PyPackageRequirements,PyUnresolvedReferences
        import IPython.kernel.zmq.ipkernel

        # noinspection PyPackageRequirements,PyUnresolvedReferences
        from IPython.kernel.zmq.ipkernel import Kernel

        # noinspection PyPackageRequirements,PyUnresolvedReferences
        from IPython.kernel.zmq.heartbeat import Heartbeat

        # noinspection PyPackageRequirements,PyUnresolvedReferences
        from IPython.kernel.zmq.session import Session

        # noinspection PyPackageRequirements,PyUnresolvedReferences
        from IPython.kernel import write_connection_file

        # noinspection PyPackageRequirements,PyUnresolvedReferences
        import zmq

        # noinspection PyPackageRequirements,PyUnresolvedReferences
        from zmq.eventloop import ioloop

        # noinspection PyPackageRequirements,PyUnresolvedReferences
        from zmq.eventloop.zmqstream import ZMQStream

        # noinspection PyPackageRequirements,PyUnresolvedReferences
        IPython.kernel.zmq.ipkernel.signal = lambda sig, f: None  # Overwrite.
    except ImportError as e:
        print("IPython import error, cannot start IPython kernel. %s" % e)
        return
    import atexit
    import socket
    import logging
    import threading

    # Do in mainthread to avoid history sqlite DB errors at exit.
    # https://github.com/ipython/ipython/issues/680
    # noinspection PyUnresolvedReferences,PyProtectedMember
    assert isinstance(threading.current_thread(), threading._MainThread)
    try:
        ip = socket.gethostbyname(socket.gethostname())
        connection_file = "ipython-kernel-%s-%s.json" % (ip, os.getpid())

        def cleanup_connection_file():
            """
            Cleanup.
            """
            try:
                os.remove(connection_file)
            except (IOError, OSError):
                pass

        atexit.register(cleanup_connection_file)

        logger = logging.Logger("IPython")
        logger.addHandler(logging.NullHandler())
        session = Session(username="kernel")

        context = zmq.Context.instance()
        transport = "tcp"
        addr = "%s://%s" % (transport, ip)
        shell_socket = context.socket(zmq.ROUTER)
        shell_port = shell_socket.bind_to_random_port(addr)
        iopub_socket = context.socket(zmq.PUB)
        iopub_port = iopub_socket.bind_to_random_port(addr)
        control_socket = context.socket(zmq.ROUTER)
        control_port = control_socket.bind_to_random_port(addr)

        hb_ctx = zmq.Context()
        heartbeat = Heartbeat(hb_ctx, (transport, ip, 0))
        hb_port = heartbeat.port
        heartbeat.start()

        shell_stream = ZMQStream(shell_socket)
        control_stream = ZMQStream(control_socket)

        kernel = Kernel(
            session=session, shell_streams=[shell_stream, control_stream], iopub_socket=iopub_socket, log=logger
        )

        write_connection_file(
            connection_file,
            shell_port=shell_port,
            iopub_port=iopub_port,
            control_port=control_port,
            hb_port=hb_port,
            ip=ip,
        )

        # print "To connect another client to this IPython kernel, use:", \
        #      "ipython console --existing %s" % connection_file
    except Exception as e:
        print("Exception while initializing IPython ZMQ kernel. %s" % e)
        return

    def ipython_thread():
        """
        IPython thread.
        """
        kernel.start()
        try:
            ioloop.IOLoop.instance().start()
        except KeyboardInterrupt:
            pass

    thread_ = threading.Thread(target=ipython_thread, name="IPython kernel")
    thread_.daemon = True
    thread_.start()


def init_cuda_not_in_main_proc_check():
    """
    Installs some hook to Theano which checks that CUDA is only used in the main proc.
    """
    # noinspection PyUnresolvedReferences,PyPackageRequirements
    import theano.sandbox.cuda as cuda

    if cuda.use.device_number is not None:
        print("CUDA already initialized in proc %i" % os.getpid())
        return
    use_original = cuda.use

    def use_wrapped(device, **kwargs):
        """
        :param device:
        :param kwargs:
        """
        print("CUDA.use %s in proc %i" % (device, os.getpid()))
        use_original(device=device, **kwargs)

    cuda.use = use_wrapped
    cuda.use.device_number = None


def debug_shell(
    user_ns: Optional[Dict[str, Any]] = None,
    user_global_ns: Optional[Dict[str, Any]] = None,
    exit_afterwards: bool = True,
):
    """
    Provides some interactive Python shell.
    Uses IPython if possible.
    Wraps to ``better_exchook.debug_shell``.

    :param user_ns:
    :param user_global_ns:
    :param exit_afterwards: will do sys.exit(1) at the end
    """
    print("Debug shell:")
    from returnn.util.basic import ObjAsDict
    from . import debug_helpers

    user_global_ns_new = dict(ObjAsDict(debug_helpers).items())
    if user_global_ns:
        user_global_ns_new.update(user_global_ns)  # may overwrite vars from DebugHelpers
    user_global_ns_new["debug"] = debug_helpers  # make this available always
    print("Available debug functions/utils (via DebugHelpers):")
    for k, v in sorted(vars(debug_helpers).items()):
        if k[:1] == "_":
            continue
        print("  %s (%s)" % (k, type(v)))
    print("Also DebugHelpers available as 'debug'.")
    if not user_ns:
        user_ns = {}
    if user_ns:
        print("Locals:")
        for k, v in sorted(user_ns.items()):
            print("  %s (%s)" % (k, type(v)))
    from returnn.util.better_exchook import debug_shell

    debug_shell(user_ns, user_global_ns_new)
    if exit_afterwards:
        print("Debug shell exit. Exit now.")
        sys.exit(1)


class PyTracer:
    """
    Trace Python function execution to get intermediate outputs from the local variables.

    E.g. for PyTorch code, when comparing results, it can be useful to see the intermediate tensors.

    Example::

        with PyTracer([my_func], torch.Tensor) as trace_my_impl:
            ...

        with PyTracer([reference_func], torch.Tensor) as trace_ref_impl:
            ...

    Or another example::

        from returnn.tensor import Tensor

        with PyTracer([my_func], Tensor) as trace_my_impl:
            ...

        with PyTracer([reference_func], torch.Tensor) as trace_ref_impl:
            ...

        check_py_traces_rf_to_pt_equal(trace_my_impl.captured_locals, trace_ref_impl.captured_locals, [...])

    See also :func:`check_py_traces_rf_to_pt_equal` to compare the traces.

    This class uses the Python :func:`sys.settrace` mechanism to trace the locals.
    It accesses ``frame.f_locals`` to get the local variables.
    Note that this behavior is slightly buggy in versions of CPython <3.13,
    see for example:
    https://github.com/python/cpython/issues/113939
    https://github.com/python/cpython/issues/74929
    And thus the behavior might be different depending on the Python version.
    In Python >=3.13, you likely get a few more locals than before.
    """

    def __init__(
        self, funcs_to_trace_list: Sequence[Union[Callable, FunctionType]], capture_type: Union[type, Tuple[type, ...]]
    ):
        """
        :param funcs_to_trace_list: list of functions to trace the locals. only those functions will be traced.
        :param capture_type: only capture variables of this type, e.g. torch.Tensor.
        """

        def _get_func_code(func: FunctionType) -> CodeType:
            while getattr(func, "__wrapped__", None) is not None:
                func = func.__wrapped__
            return func.__code__

        self.funcs_to_trace_list = funcs_to_trace_list
        self._code_obj_to_func = {_get_func_code(func): func for func in self.funcs_to_trace_list}
        self.capture_type = capture_type

        self._prev_trace_func = None
        self.captured_locals = {}  # func -> (list of calls) -> tensor local name -> (list of versions) -> tensor

    def __enter__(self) -> PyTracer:
        self._prev_trace_func = sys.gettrace()
        # Note: We might get such a warning here when using PyDev/PyCharm:
        #   PYDEV DEBUGGER WARNING:
        #   sys.settrace() should not be used when the debugger is being used.
        #   This may cause the debugger to stop working correctly.
        # This is ok.
        # In fact, we even call this prev_trace_func in our own __call__ as well,
        # so debugging (breakpoints etc.) should still work (I tested it.).
        sys.settrace(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.settrace(self._prev_trace_func)
        self._prev_trace_func = None

    def __call__(self, frame: FrameType, event, arg) -> Optional[PyTracer]:
        """
        Trace func to get intermediate outputs.
        """
        prev_trace_func_res = None
        if self._prev_trace_func:
            prev_trace_func_res = self._prev_trace_func(frame, event, arg)
        func = self._code_obj_to_func.get(frame.f_code)
        if func:
            if event == "call":
                self.captured_locals.setdefault(func, []).append({})
            else:
                # Note that accessing frame.f_locals is not always totally accurate.
                # See the corresponding comment in our class docstring.
                for k, v in frame.f_locals.items():
                    if not isinstance(v, self.capture_type):
                        continue
                    prev = self.captured_locals[func][-1].get(k, None)
                    if prev is None or prev[-1] is not v:
                        var_ls = self.captured_locals[func][-1].setdefault(k, [])
                        print(
                            f"{func.__qualname__}[{len(self.captured_locals[func]) - 1}]"
                            f" ({type(v).__qualname__}) {k}[{len(var_ls)}] = {v}"
                        )
                        var_ls.append(v)
            return self
        return prev_trace_func_res


def check_py_traces_rf_to_pt_equal(
    trace_rf: Dict[Callable, List[Dict[str, List[Tensor]]]],
    trace_pt: Dict[Callable, List[Dict[str, List[torch.Tensor]]]],
    checks: List[
        Tuple[
            Tuple[Callable, int, str, int],
            Tuple[Callable, int, str, int],
            Union[Tuple[Union[Dim, str], ...], Callable[[torch.Tensor], Tensor]],
        ],
    ],
):
    """
    Compares traces from some RETURNN-frontend (RF) based implementation
    with some pure PyTorch (PT) based implementation.

    :param trace_rf: RETURNN-frontend trace, from :class:`PyTracer`
    :param trace_pt: pure PyTorch trace, from :class:`PyTracer`
    :param checks: list of checks to perform. each check is a tuple of:
        - RF trace entry, e.g. (func, i, name, j)
        - PT trace entry, e.g. (func, i, name, j)
        - PT dims, e.g. (batch_dim, other_dim, ...).
            Instead of Dim, you can also use a string, which will be resolved from the RF trace
            (then you also need ``Dim`` in ``capture_type`` of the :class:`PyTracer`).
            If callable, it gets the PyTorch tensor and should return the RETURNN tensor.
            Sometimes you might want to perform some reshaping, slicing, or similar,
            and then use rf.convert_to_tensor.
    """
    import random
    import torch
    from returnn.tensor import Tensor, Dim
    import returnn.frontend as rf

    # noinspection PyProtectedMember
    from torch.testing._comparison import make_tensor_mismatch_msg

    dummy_forward_compat_kwargs = {f"_random_forward_compat_arg_{random.randint(0, 1_000_000)}": "please"}

    def _get_entry(trace, func, i, name, j):
        return trace[func][i][name][j]

    def _resolve_dim(dim: Union[Dim, str]) -> Dim:
        if isinstance(dim, Dim):
            return dim
        elif isinstance(dim, str):
            dim = _get_entry(trace_rf, *check_rf[:2], dim, -1)
            assert isinstance(dim, Dim)
            return dim
        else:
            raise TypeError(f"invalid dim type: {dim!r}")

    def _format_check(check: Tuple[Union[FunctionType, Callable], int, str, int]) -> str:
        func, i, var_name, j = check
        return f"{func.__qualname__}[{i}] {var_name}[{j}]"

    non_matching = []
    for check_rf, check_pt, pt_dims in checks:
        print(f"checking {_format_check(check_rf)} vs {_format_check(check_pt)} ({pt_dims})...")
        tensor_rf: Tensor = _get_entry(trace_rf, *check_rf)
        tensor_pt: torch.Tensor = _get_entry(trace_pt, *check_pt)
        if callable(pt_dims):
            tensor_pt_ = pt_dims(tensor_pt, name=check_pt[2], resolve_dim=_resolve_dim, **dummy_forward_compat_kwargs)
        else:
            pt_dims = [_resolve_dim(dim) for dim in pt_dims]
            tensor_pt_ = rf.convert_to_tensor(tensor_pt, dims=pt_dims, name=check_pt[2])
        tensor_pt_ = tensor_pt_.copy_transpose(tensor_rf.dims)
        tensor_rf = tensor_rf.copy_masked(0.0)
        tensor_pt_ = tensor_pt_.copy_masked(0.0)
        rtol, atol = 1e-4, 1e-5  # defaults from torch.testing.assert_allclose
        matches = torch.isclose(tensor_rf.raw_tensor, tensor_pt_.raw_tensor, rtol=rtol, atol=atol)
        if matches.all():
            print("  matched!")
        else:
            msgs = make_tensor_mismatch_msg(
                tensor_rf.raw_tensor, tensor_pt_.raw_tensor, matches, rtol=rtol, atol=atol, identifier="Tensors"
            )
            msgs_prefix = [
                f"{check_rf} vs {check_pt}:",
                f"  RF: {tensor_rf} {tensor_rf.raw_tensor}",
                f"  PT: {tensor_pt_} {tensor_pt_.raw_tensor}",
            ]
            msgs = ["  " + line for line in msgs.splitlines() if line and "Tensors are not " not in line]
            indices = (~matches).nonzero().detach().cpu()
            for idx in indices[:5]:
                idx = tuple(idx.numpy())
                msgs.append(f"  non-matching at {idx}: RF={tensor_rf.raw_tensor[idx]} PT={tensor_pt_.raw_tensor[idx]}")
            if len(indices) > 5:
                msgs.append("  non-matching ...")
            non_matching.append("\n".join(msgs_prefix + msgs))
            print(f"  mismatch!")
            for msg in msgs:
                print(msg)

    if non_matching:
        raise AssertionError("\n\n".join(non_matching))



"""
Utilities for dimension tags, dimensions, axes.
"""

from __future__ import annotations
from typing import Optional, Union, TypeVar, Sequence, Tuple
from returnn.tensor import Tensor, Dim
import returnn.frontend as rf
from ._backend import get_backend_by_tensor, global_backend

T = TypeVar("T")

__all__ = [
    "range_over_dim",
    "range_over_dim_strided",
    "range_over_merged_dims",
    "replace_dim",
    "replace_dim_v2",
    "set_sparse_dim",
    "dim_match_priority_when_needed",
    "num_elements_of_shape",
    "masked_fraction_of_shape",
    "last_frame_position_of_dim",
]


def range_over_dim(dim: Dim, *, dtype: Optional[str] = None, device: Optional[str] = None) -> Tensor[T]:
    """
    :param dim:
    :param dtype:
    :param device,
    :return: tensor with shape [dim]
    """
    if dim.dyn_size_ext is not None:
        backend = get_backend_by_tensor(dim.dyn_size_ext, fallback=global_backend)
    else:
        backend = global_backend
    return backend.range_over_dim(dim, dtype=dtype, device=device)


def range_over_dim_strided(
    dim: Dim,
    *,
    stride: Union[int, Tensor],
    out_dim: Optional[Dim] = None,
    dtype: Optional[str] = None,
    device: Optional[str] = None,
) -> Tuple[Tensor[T], Dim]:
    """
    :param dim:
    :param stride:
    :param out_dim:
    :param dtype:
    :param device,
    :return: tensor with shape [dim], out_dim
    """
    if out_dim is None:
        out_dim = dim.ceildiv_right(stride)
    return rf.range_over_dim(out_dim, dtype=dtype, device=device) * stride, out_dim


def range_over_merged_dims(
    dims: Sequence[Dim], *, dtype: Optional[str] = None, device: Optional[str] = None
) -> Tensor[T]:
    """
    This is if you want to index into a merged dim.
    Related: :func:`rf.merge_dims`.

    :param dims:
    :param dtype:
    :param device:
    :return: tensor with shape [dim_0, ..., dim_n] -> sparse_dim = merged_dim, where merged_dim = dim_0 * ... * dim_n
    """
    assert len(dims) >= 1
    merged_dim = dims[0]
    for dim in dims[1:]:
        merged_dim *= dim
    indices = rf.range_over_dim(merged_dim, dtype=dtype, device=device)
    if len(dims) > 1:
        indices = rf.split_dims(indices, axis=merged_dim, dims=dims)
    return indices


def replace_dim(source: Tensor, *, in_dim: Dim, out_dim: Optional[Dim] = None) -> Tuple[Tensor, Dim]:
    """
    Also see: :func:`replace_dim_v2`, :func:`rf.merge_dims`, :func:`rf.split_dims`.

    :param source:
    :param in_dim: should be in ``source.dims``, to be replaced.
        If you want to replace the ``source.sparse_dim``, see :func:`set_sparse_dim`.
    :param out_dim: If not given, will create a new dim with the same size as ``in_dim``.
        Note: If the size of ``out_dim`` is different from ``in_dim``,
        currently the dim tag is replaced and there is no error -- this is not checked.
    :return: ``source`` with ``in_dim`` replaced by ``out_dim``, and ``out_dim``.
    """
    if not out_dim:
        out_dim = in_dim.copy(same_as_self=False, description="new-dim")
    # noinspection PyProtectedMember
    return source._raw_backend.replace_dim(source, in_dim=in_dim, out_dim=out_dim), out_dim


def replace_dim_v2(
    source: Tensor, *, in_dim: Dim, out_dim: Dim, allow_expand: bool = True, allow_shrink: bool = True
) -> Tensor:
    """
    Extends :func:`replace_dim` by also allowing to expand or shrink the dim
    (or rather, to not ignore this; when :func:`replace_dim` is used on a dim with different size,
     it will ignore this and anyway accept the new dim tag (currently)).

    :param source:
    :param in_dim: should be in ``source.dims``, to be replaced.
        If you want to replace the ``source.sparse_dim``, see :func:`set_sparse_dim`.
    :param out_dim: should not be in ``source.dims``, to be replaced.
        Note: In contrast to :func:`replace_dim`, you must provide this explicitly.
    :param allow_expand: if True, allow to expand the dim, i.e. if ``out_dim.size > in_dim.size``.
    :param allow_shrink: if True, allow to shrink the dim, i.e. if ``out_dim.size < in_dim.size``.
    :return: ``source`` with ``in_dim`` replaced by ``out_dim``.
    """
    if not rf.is_executing_eagerly():
        raise NotImplementedError  # just not implemented yet. we can do via :func:`cond`
    if in_dim not in source.dims:
        raise ValueError(f"replace_dim_v2: dim {in_dim} not in {source}")
    if out_dim in source.dims:
        raise ValueError(f"replace_dim_v2: dim {out_dim} already in {source}")
    old_size = in_dim.get_dim_value()
    new_size = out_dim.get_dim_value()
    if new_size == old_size:
        res, _ = rf.replace_dim(source, in_dim=in_dim, out_dim=out_dim)
    elif new_size > old_size:
        if not allow_expand:
            raise ValueError(
                f"replace_dim_v2: not allowed to expand: {old_size} -> {new_size},"
                f" for {in_dim=} {out_dim=}, in {source=}"
            )
        res, _ = rf.pad(
            source,
            axes=[in_dim],
            padding=[(0, out_dim.get_dim_value_tensor() - in_dim.get_dim_value_tensor())],
            out_dims=[out_dim],
            value=0,
        )
    else:
        if not allow_shrink:
            raise ValueError(
                f"replace_dim_v2: not allowed to shrink: {old_size} -> {new_size},"
                f" for {in_dim=} {out_dim=}, in {source=}"
            )
        res, _ = rf.slice(source, axis=in_dim, size=out_dim)
    return res


def set_sparse_dim(source: Tensor, sparse_dim: Dim) -> Tensor:
    """
    :param source:
    :param sparse_dim:
    :return: source with sparse_dim set
    """
    # noinspection PyProtectedMember
    return source._raw_backend.set_sparse_dim(source, sparse_dim)


def dim_match_priority_when_needed(dim: Dim, *other_dims: Dim) -> Dim:
    """
    :return: maybe copy of dim with higher match_priority if needed to distinguish from other_dims

    Why or when is this needed?

    For activation values, this should never be needed,
    and all dims should be unique.

    In case of self-attention, the standard way is to create a separate distinct dim
    to perform the attention reduction over.
    See :class:`SelfAttention`.

    However, in case of weight matrices, it is not unusual to have the same dim for both the input and output,
    so a square weight matrix.
    When reduction is performed in :func:`matmul`, we want to match the input feature dim
    to the dim in the weight matrix with higher match priority.

    So :func:`dim_match_priority_when_needed` would be applied on the input feature dim.

    https://github.com/rwth-i6/returnn/pull/871
    https://github.com/rwth-i6/returnn_common/issues/17#issuecomment-997463222
    """
    if dim in other_dims:
        return dim.copy(match_priority=1)
    return dim


def num_elements_of_shape(
    dims: Union[Dim, Sequence[Dim]], *, use_mask: bool = True, device: Optional[str] = None
) -> Union[int, Tensor]:
    """
    :param dims:
    :param use_mask:
    :param device: only for the case when we return a Tensor. by default, this is CPU (just as the size tensor).
    :return: num elements of a tensor of shape dims, properly considering masking
    """
    if isinstance(dims, Dim):
        dims = [dims]
    if not use_mask:
        n = 1
        for dim in dims:
            n *= dim.get_dim_value_tensor()
        return n
    if all(dim.is_static() for dim in dims):
        n = 1
        for dim in dims:
            n *= dim.size
        return n
    if len(dims) == 1:
        return dims[0].get_size_tensor(device=device)

    n: Union[int, Tensor] = 1
    postponed_dims = []
    for i, dim in enumerate(dims):
        # E.g. if dim==B, and some other dim dyn_size_ext has B, then we need to postpone this.
        related_dims = []
        for j, dim_ in enumerate(dims):
            if i == j:
                continue
            if dim_.dyn_size_ext is not None and dim in dim_.dyn_size_ext.dims:
                related_dims.append(dim_)
        if not related_dims:
            if dim.is_static():
                n *= dim.size
            else:
                n *= dim.get_size_tensor(device=device)
        else:
            postponed_dims.append(dim)
    if postponed_dims:
        n: Tensor
        n = rf.reduce_sum(n, axis=postponed_dims)
    return n


def masked_fraction_of_shape(dims: Union[Dim, Sequence[Dim]], *, inverse: bool = False) -> Union[int, float, Tensor]:
    """
    :param dims:
    :param inverse: if True, return the inverse of the fraction
    :return: :func:`num_elements_of_shape`(dims) / prod(dims) if not inverse else prod(dims) / num_elements
    """
    if isinstance(dims, Dim):
        dims = [dims]
    if not any(dim.need_masking() for dim in dims):
        return 1
    num_elems_masked = num_elements_of_shape(dims)
    num_elems_total = 1
    for dim in dims:
        num_elems_total *= dim.get_dim_value_tensor()
    return (num_elems_masked / num_elems_total) if not inverse else (num_elems_total / num_elems_masked)


def last_frame_position_of_dim(
    dim: Dim, *, device: Optional[str] = None, allow_scalar_on_cpu: bool = True
) -> Union[int, Tensor]:
    """
    :param dim:
    :param device:
    :param allow_scalar_on_cpu: if device is not given, we would use rf.get_default_device() for dynamic sizes.
        If this is True, we would allow to return a scalar dynamic size on CPU.
    :return: last frame position of dim
    """
    if dim.size is not None:
        if dim.size > 0:
            return dim.size - 1
        return 0
    if device is None:
        if allow_scalar_on_cpu and not dim.dyn_size_ext.dims:
            device = "cpu"
        else:
            device = rf.get_default_device()
    pos = dim.get_dyn_size_ext_for_device(device) - 1
    pos = rf.maximum(pos, 0)
    pos.sparse_dim = dim
    return pos



"""
Piecewise linear function
"""

from __future__ import annotations
from typing import Union, Dict
import numpy as np
from returnn.tensor import Tensor, Dim
import returnn.frontend as rf


__all__ = ["PiecewiseLinear"]


class PiecewiseLinear(rf.Module):
    """
    Piecewise linear function.
    """

    def __init__(self, points: Dict[Union[int, float], Union[float, Tensor]]):
        """
        :param points: dict of key -> value pairs.
        """
        super().__init__()
        if not points:
            raise ValueError(f"{self}: points must not be empty")
        self._points_sorted = sorted(points.items())
        self.points_dim = Dim(len(self._points_sorted), name="pcw_schd_pieces")
        # Note: Use rf.Parameter to work around deepcopy issue. https://github.com/rwth-i6/returnn/issues/1541
        self._keys = rf.Parameter(
            rf.convert_to_tensor(
                np.array([k for k, _ in self._points_sorted], dtype=rf.get_default_float_dtype()),
                dims=[self.points_dim],
            ),
            auxiliary=True,
        )
        self._values = rf.Parameter(
            rf.stack([rf.convert_to_tensor(v) for _, v in self._points_sorted], out_dim=self.points_dim)[0],
            auxiliary=True,
        )

    def __call__(self, x: Tensor) -> Tensor:
        """
        :param x: (x_dims...) -> value in keys
        :return: y: (x_dims...,y_dims...) -> value in values
        """
        index = rf.search_sorted(self._keys, x, axis=self.points_dim)
        index = rf.clip_by_value(index, 1, self.points_dim.dimension - 1)
        x_start = rf.gather(self._keys, indices=index - 1)
        x_end = rf.gather(self._keys, indices=index)
        x_frac = (x - x_start) / (x_end - x_start)
        x_frac = rf.clip_by_value(x_frac, 0.0, 1.0)
        y_start = rf.gather(self._values, indices=index - 1)
        y_end = rf.gather(self._values, indices=index)
        return rf.lerp(y_start, y_end, x_frac)



"""
Provides :class:`MapDatasetBase`
"""

from __future__ import annotations
from typing import Any, List, Tuple, Dict
from returnn.datasets.basic import DatasetSeq
from returnn.datasets.cached2 import CachedDataset2
from returnn.util.basic import OptionalNotImplementedError


class MapDatasetBase:
    """
    This dataset can be used as template to implement user-side Datasets,
    where the data can be access in arbitrary order.
    For global sorting, the length information needs to be known beforehand, see get_seq_len.
    """

    def __init__(self, data_types=None):
        """
        :param dict[str,dict] data_types: data_key -> constructor parameters of Data object, for all data streams the
          dataset provides (inputs and targets). E.g. {'data': {'dim': 1000, 'sparse': True, ...}, 'classes': ...}.
        """
        self.data_types = data_types or {}

    def __len__(self):
        """
        :return: total number of sequences in the dataset
        :rtype: int
        """
        raise NotImplementedError

    def __getitem__(self, seq_idx):
        """
        This function does the actual data loading, the order can be arbitrary.

        :param int seq_idx:
        :return: The content of a single dataset entry
        :rtype: dict[str,numpy.array]
        """
        raise NotImplementedError

    def get_seq_len(self, seq_idx):
        """
        This optional function provides the sequence length for the `seq_ordering` parameter.
        If not specified only a limited set of options is available.

        :param int seq_idx:
        :return: sequence length
        :rtype: int
        """
        raise OptionalNotImplementedError

    def get_seq_tag(self, seq_idx):
        """
        :param int seq_idx:
        :return: tag for the sequence of the given index, default is 'seq-{seq_idx}'.
        :rtype: str
        """
        return "seq-%i" % seq_idx

    def get_seq_order(self, epoch=None):
        """
        Override to implement a dataset specific sequence order for a given epoch.
        The number of sequences can be less than the total number.
        This will override the effects of `partition_epoch` and `seq_ordering` when using MapDatasetWrapper.

        :param int epoch:
        :return: sequence order (list of sequence indices)
        :rtype: list[int]
        """
        raise OptionalNotImplementedError


class MapDatasetWrapper(CachedDataset2):
    """
    Takes a MapDataset and turns it into a returnn.datasets.Dataset by providing the required class methods.
    """

    def __init__(self, map_dataset, **kwargs):
        """
        :param MapDatasetBase|function map_dataset: the MapDataset to be wrapped
        """
        super(MapDatasetWrapper, self).__init__(**kwargs)

        if callable(map_dataset):
            map_dataset = map_dataset()
        assert isinstance(map_dataset, MapDatasetBase)

        self._dataset = map_dataset
        self._seq_order = None
        assert map_dataset.data_types, f"{self}: map_dataset {map_dataset} needs to provide data_types"
        self.num_outputs = {key: _get_num_outputs_entry(key, opts) for key, opts in map_dataset.data_types.items()}

    @property
    def map_dataset(self) -> MapDatasetBase:  # also important for Dataset.__reduce__
        """
        :return: the wrapped MapDataset
        """
        return self._dataset

    @property
    def num_seqs(self):
        """
        :returns number of sequences in the current epoch
        :rtype: int
        """
        if self._seq_order is None:
            raise NotImplementedError("'num_seqs' is only known after calling init_seq_order().")
        return len(self._seq_order)

    def get_total_num_seqs(self, *, fast: bool = False) -> int:
        """
        :return: total number of seqs
        """
        return len(self._dataset)

    def init_seq_order(self, epoch=None, seq_list=None, seq_order=None):
        """
        :param int|None epoch:
        :param list[str]|None seq_list: List of sequence tags, to set a predefined order.
        :param list[int]|None seq_order: List of corpus sequence indices, to set a predefined order.
        :rtype: bool
        :returns whether the order changed (True is always safe to return)
        """
        super(MapDatasetWrapper, self).init_seq_order(epoch=epoch, seq_list=seq_list, seq_order=seq_order)

        if seq_list is not None:
            assert seq_order is None
            tag_to_idx = {
                self._dataset.get_seq_tag(corpus_seq_idx): corpus_seq_idx
                for corpus_seq_idx in range(len(self._dataset))
            }
            seq_order = [tag_to_idx[tag] for tag in seq_list]
        if seq_order is not None:
            self._seq_order = seq_order
        else:
            try:
                self._seq_order = self._dataset.get_seq_order(epoch=epoch)
            except OptionalNotImplementedError:
                try:
                    self._seq_order = self.get_seq_order_for_epoch(
                        epoch=epoch, num_seqs=len(self._dataset), get_seq_len=self._dataset.get_seq_len
                    )
                except OptionalNotImplementedError:
                    # only support seq_ordering that need no length here
                    assert self.seq_ordering in ["default", "reverse", "random"], (
                        f"{self}: dataset {self._dataset}.get_seq_len is not implemented,"
                        f" seq_ordering {self.seq_ordering!r} is not supported"
                    )
                    self._seq_order = self.get_seq_order_for_epoch(
                        epoch=epoch, num_seqs=len(self._dataset), get_seq_len=None
                    )

        return True

    def supports_seq_order_sorting(self) -> bool:
        """supports sorting"""
        try:
            self._dataset.get_seq_len(0)
            return True
        except OptionalNotImplementedError:
            return False

    def _collect_single_seq(self, seq_idx):
        """
        :param int seq_idx: sorted seq idx
        :return:
        """
        corpus_seq_idx = self.get_corpus_seq_idx(seq_idx)
        return DatasetSeq(
            seq_idx, features=self._dataset[corpus_seq_idx], seq_tag=self._dataset.get_seq_tag(corpus_seq_idx)
        )

    def get_current_seq_order(self):
        """
        :rtype: typing.Sequence[int]
        """
        assert self._seq_order is not None
        return self._seq_order

    def get_tag(self, sorted_seq_idx):
        """
        :param sorted_seq_idx:
        :return:
        """
        seq_tag = self._dataset.get_seq_tag(self.get_corpus_seq_idx(sorted_seq_idx))
        return seq_tag

    def get_all_tags(self) -> List[str]:
        """
        :return: list of all tags
        """
        return [self._dataset.get_seq_tag(i) for i in range(len(self._dataset))]

    def get_corpus_seq_idx(self, sorted_seq_idx):
        """
        :param int sorted_seq_idx:
        :return corpus_seq_idx
        :rtype: int
        """
        return self._seq_order[sorted_seq_idx]

    def have_corpus_seq_idx(self):
        """
        :rtype: bool
        :return: whether you can call self.get_corpus_seq_idx()
        """
        return True

    def get_data_keys(self) -> List[str]:
        """
        :return: keys
        """
        return list(self._dataset.data_types.keys())

    def get_data_dim(self, key):
        """
        :param str key: e.g. "data" or "classes"
        :return: number of classes, no matter if sparse or not
        :rtype: int
        """
        if key in self._dataset.data_types:
            return self._dataset.data_types[key].get("dim", 1)
        return 1  # unknown

    def get_data_dtype(self, key):
        """
        :param str key: e.g. "data" or "classes"
        :return: dtype as str, e.g. "int32" or "float32"
        :rtype: str
        """
        if key in self._dataset.data_types and "dtype" in self._dataset.data_types[key]:
            return self._dataset.data_types[key]["dtype"]
        if self.is_data_sparse(key):
            return "int32"
        return "float32"

    def is_data_sparse(self, key):
        """
        :param str key: e.g. "data" or "classes"
        :return: whether the data is sparse
        :rtype: bool
        """
        # Note: We cannot call get_data_dtype, as we would maybe result in infinite recursion.
        if key in self._dataset.data_types:
            return self._dataset.data_types[key].get("sparse", False)
        return False

    def get_data_shape(self, key):
        """
        :returns get_data(*, key).shape[1:], i.e. num-frames excluded
        :rtype: list[int]
        """
        if key in self._dataset.data_types:
            if "shape" in self._dataset.data_types[key].keys():
                if self._dataset.data_types[key]["shape"][0] is None:
                    return self._dataset.data_types[key]["shape"][1:]
                else:
                    assert False, "data shape has no time axis, calling get_data_shape is not possible"
        if self.is_data_sparse(key):
            return []
        return [self.get_data_dim(key)]


class FromListDataset(MapDatasetBase):
    """
    Simple implementation of a MapDataset where all data is given in a list.
    """

    def __init__(self, data_list, sort_data_key=None, **kwargs):
        """
        :param list[dict[str,numpy.ndarray]] data_list: sequence data as a dict data_key -> data for all sequences.
        :param str sort_data_key: Sequence length will be determined from data of this data_key.
        """
        self._data_list = data_list
        self._sort_data_key = sort_data_key
        super(FromListDataset, self).__init__(**kwargs)

    def __len__(self):
        """
        :return: total number of sequences in the dataset
        :rtype: int
        """
        return len(self._data_list)

    def __getitem__(self, seq_idx):
        """
        :param int seq_idx:
        :return: The content of a single dataset entry
        :rtype dict[str,numpy.array]
        """
        return self._data_list[seq_idx]

    def get_seq_len(self, seq_idx):
        """
        :param seq_idx:
        :return: length of data for 'sort_data_key'
        :rtype: int
        """
        assert self._sort_data_key, "Specify which data key should be used for sequence sorting via 'sort_data_key'."
        return len(self._data_list[seq_idx][self._sort_data_key])


def _get_num_outputs_entry(name: str, opts: Dict[str, Any]) -> Tuple[int, int]:
    """
    :param opts: data opts from data_types in MapDatasetBase
    :return: num_outputs entry: (num-classes, len(shape))

    This is maybe optional at some point when we remove num_outputs in Dataset.
    """
    from returnn.tensor import Tensor

    data = Tensor(name, **opts)
    return data.dim or (data.shape[-1] if data.shape else 0), len(data.shape)
